# Copyright 2025 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Genomics intervals used for training model folds."""

import enum

from alphagenome.models import dna_client
import immutabledict
import pandas as pd


_DEFAULT_EXAMPLE_REGIONS = immutabledict.immutabledict({
    dna_client.Organism.HOMO_SAPIENS: (
        'https://github.com/calico/borzoi/raw/'
        '5c9358222b5026abb733ed5fb84f3f6c77239b37/data/sequences_human.bed.gz'
    ),
    dna_client.Organism.MUS_MUSCULUS: (
        'https://github.com/calico/borzoi/raw/'
        '5c9358222b5026abb733ed5fb84f3f6c77239b37/data/sequences_mouse.bed.gz'
    ),
})


class Subset(enum.Enum):
  """Subset of the data."""

  TRAIN = 0
  VALID = 1
  TEST = 2


# Fold ONE is aligned with all trained Borzoi checkpoints: 3 and 4 are held out.
_VALID_FOLD = immutabledict.immutabledict({
    0: 'fold0',
    1: 'fold3',
    2: 'fold2',
    3: 'fold6',
    -1: 'fold0',
})

_TEST_FOLD = immutabledict.immutabledict({
    0: 'fold1',
    1: 'fold4',
    2: 'fold5',
    3: 'fold7',
    -1: 'fold1',
})

_MODEL_VERSION_TO_FOLD = immutabledict.immutabledict({
    dna_client.ModelVersion.FOLD_0: 0,
    dna_client.ModelVersion.FOLD_1: 1,
    dna_client.ModelVersion.FOLD_2: 2,
    dna_client.ModelVersion.FOLD_3: 3,
    dna_client.ModelVersion.ALL_FOLDS: -1,
})


def get_all_folds() -> list[str]:
  """Returns the names of all data folds."""
  return [f'fold{i}' for i in range(8)]


def get_fold_names(
    model_version: dna_client.ModelVersion, subset: Subset
) -> list[str]:
  """Returns the data folds used for the model version."""
  match subset:
    case Subset.VALID:
      return [_VALID_FOLD[_MODEL_VERSION_TO_FOLD[model_version]]]
    case Subset.TEST:
      return [_TEST_FOLD[_MODEL_VERSION_TO_FOLD[model_version]]]
    case Subset.TRAIN:
      all_folds = get_all_folds()
      if _MODEL_VERSION_TO_FOLD[model_version] == -1:
        return all_folds
      remove_folds = get_fold_names(
          model_version, Subset.VALID
      ) + get_fold_names(model_version, Subset.TEST)
      for fold in remove_folds:
        all_folds.remove(fold)
      return all_folds
    case _:
      raise ValueError(f'Unknown {subset=}')


def get_fold_intervals(
    model_version: dna_client.ModelVersion,
    organism: dna_client.Organism,
    subset: Subset,
    example_regions_path: str | None = None,
) -> pd.DataFrame:
  """Returns the training intervals for the model version."""
  if example_regions_path is None:
    example_regions_path = _DEFAULT_EXAMPLE_REGIONS[organism]

  example_regions = pd.read_csv(
      example_regions_path,
      sep='\t',
      names=['chromosome', 'start', 'end', 'fold'],
  )
  return example_regions[
      example_regions.fold.isin(get_fold_names(model_version, subset))
  ]



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from alphagenome.protos import dna_model_pb2
from alphagenome.data import genome
import numpy as np


class StrandTest(parameterized.TestCase):

  @parameterized.parameters(
      list(
          zip(
              genome.STRAND_OPTIONS,
              (
                  genome.Strand.POSITIVE,
                  genome.Strand.NEGATIVE,
                  genome.Strand.UNSTRANDED,
              ),
          )
      )
  )
  def test_from_str(self, strand_str, expected):
    strand = genome.Strand.from_str(strand_str)
    self.assertEqual(strand, expected)

  def test_invalid_strand_str(self):
    with self.assertRaises(ValueError):
      genome.Strand.from_str('foo')

  @parameterized.parameters(
      (genome.Strand.POSITIVE,),
      (genome.Strand.NEGATIVE,),
      (genome.Strand.UNSTRANDED,),
  )
  def test_from_proto(self, strand):
    proto = genome.Strand.to_proto(strand)
    round_trip = genome.Strand.from_proto(proto)
    self.assertEqual(strand, round_trip)

  def test_invalid_strand_proto(self):
    with self.assertRaises(ValueError):
      genome.Strand.from_proto(dna_model_pb2.Strand.STRAND_UNSPECIFIED)


class IntervalTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(chromosome='chr1', start=0, end=100, strand='-'),
      dict(chromosome='chr2', start=100, end=200, strand='.'),
      dict(chromosome='chr3', start=100, end=200, strand='+'),
  )
  def test_proto_roundtrip(self, chromosome, start, end, strand):
    interval = genome.Interval(chromosome, start, end, strand)
    proto = interval.to_proto()
    round_trip = genome.Interval.from_proto(proto)
    self.assertEqual(interval, round_trip)

  def test_init_valid(self):
    genome.Interval('chr1', 10, 10)
    genome.Interval(
        'chr1',
        10,
        20,
        name='test',
        info=dict(a=1, b='b', c=[1, 2], d=dict(e=1)),
    )

  def test_init_raises_value_error(self):
    with self.assertRaises(ValueError):
      genome.Interval('chr1', 5, 1)  # start < end

  def test_attributes(self):
    interval = genome.Interval('chr1', 10, 20, strand='-')
    self.assertEqual(interval.start, 10)
    self.assertEqual(interval.end, 20)
    self.assertEqual(interval.chromosome, 'chr1')
    self.assertEqual(interval.name, '')
    self.assertIsInstance(interval.info, dict)
    self.assertEmpty(interval.info)
    interval.info['test'] = 10
    self.assertEqual(interval.info['test'], 10)
    self.assertTrue(interval.negative_strand)

    self.assertEqual(interval.width, 10)

  def test_copy(self):
    # make sure the original got unchanged
    interval = genome.Interval('chr1', 10, 20, strand='-')
    i2 = interval.copy()
    i2.info['test'] = 10
    self.assertEqual(i2.info['test'], 10)
    self.assertEmpty(interval.info)

  def test_serialization(self):
    interval = genome.Interval('chr1', 0, 100, '-')
    self.assertEqual(interval, genome.Interval.from_str('chr1:0-100:-'))
    self.assertEqual(str(interval), 'chr1:0-100:-')

  def test_serialization_no_strand(self):
    # Serialized intervals are allowed to be missing a strand, in which case
    # they are treated as unstranded. Serializing an interval always adds the
    # strand.
    interval = genome.Interval('chr1', 0, 100)
    self.assertEqual(interval, genome.Interval.from_str('chr1:0-100'))
    self.assertEqual(interval, genome.Interval.from_str('chr1:0-100:.'))
    self.assertEqual(str(interval), 'chr1:0-100:.')

  def test_from_to_string(self):
    interval = genome.Interval('chr1', 10, 20, strand='-')
    self.assertIsInstance(str(interval), str)
    self.assertEqual(interval, genome.Interval.from_str(str(interval)))

  def test_from_to_string_negative_start(self):
    interval = genome.Interval('chr1', -10, 20, strand='-')
    self.assertIsInstance(str(interval), str)
    self.assertEqual(interval, genome.Interval.from_str(str(interval)))

  def test_from_to_string_negative_end(self):
    interval = genome.Interval('chr1', -20, -10, strand='-')
    self.assertIsInstance(str(interval), str)
    self.assertEqual(interval, genome.Interval.from_str(str(interval)))

  def test_overlaps_overlapping_at_start(self):
    self.assertTrue(
        genome.Interval('chr1', 1, 5).overlaps(genome.Interval('chr1', 3, 7))
    )

  def test_overlaps_overlapping_at_end(self):
    self.assertTrue(
        genome.Interval('chr1', 1, 5).overlaps(genome.Interval('chr1', 0, 2))
    )

  def test_overlaps_overlapping_contains(self):
    self.assertTrue(
        genome.Interval('chr1', 1, 5).overlaps(genome.Interval('chr1', 0, 6))
    )

  def test_overlaps_nonoverlapping_at_start(self):
    self.assertFalse(
        genome.Interval('chr1', 1, 5).overlaps(genome.Interval('chr1', 0, 1))
    )

  def test_overlaps_nonoverlapping_at_end(self):
    self.assertFalse(
        genome.Interval('chr1', 1, 5).overlaps(genome.Interval('chr1', 5, 6))
    )

  def test_overlaps_nonoverlapping_different_chromosome(self):
    self.assertFalse(
        genome.Interval('chr1', 1, 5).overlaps(genome.Interval('chr2', 0, 6))
    )

  def test_contains_small_within_larger(self):
    self.assertTrue(
        genome.Interval('chr1', 1, 5).contains(genome.Interval('chr1', 2, 3))
    )

  def test_contains_large_within_small(self):
    self.assertFalse(
        genome.Interval('chr1', 2, 3).contains(genome.Interval('chr1', 1, 5))
    )

  def test_contains_same_size(self):
    self.assertTrue(
        genome.Interval('chr1', 1, 5).contains(genome.Interval('chr1', 1, 5))
    )

  def test_contains_overlapping_end(self):
    self.assertFalse(
        genome.Interval('chr1', 1, 5).contains(genome.Interval('chr1', 4, 9))
    )

  def test_contains_overlapping_start(self):
    self.assertFalse(
        genome.Interval('chr1', 4, 10).contains(genome.Interval('chr1', 1, 5))
    )

  def test_contains_non_overlapping_end(self):
    self.assertFalse(
        genome.Interval('chr1', 1, 5).contains(genome.Interval('chr1', 6, 9))
    )

  def test_contains_non_overlapping_start(self):
    self.assertFalse(
        genome.Interval('chr1', 6, 9).contains(genome.Interval('chr1', 1, 5))
    )

  def test_contains_different_chromosome(self):
    self.assertFalse(
        genome.Interval('chr1', 1, 5).contains(genome.Interval('chr2', 1, 5))
    )

  def test_shift_dont_use_strand(self):
    interval = genome.Interval('chr1', 10, 20, strand='-')
    i2 = interval.shift(10, use_strand=False)

    # original unchanged
    self.assertEqual(interval.start, 10)
    self.assertEqual(interval.end, 20)

    self.assertEqual(i2.start, 20)
    self.assertEqual(i2.end, 30)

  def test_shift_use_strand(self):
    interval = genome.Interval('chr1', 10, 20, strand='-')
    i2 = interval.shift(10)  # use_strand = True by default.
    self.assertEqual(i2.start, 0)
    self.assertEqual(i2.end, 10)

    # Shift outside of the reference.
    self.assertFalse(interval.shift(20, use_strand=True).within_reference())

    i2 = interval.shift(15, use_strand=True).truncate()
    self.assertEqual(i2.start, 0)
    self.assertEqual(i2.end, 5)

    self.assertEqual(interval.center(), 15)

  def test_pad_does_not_mutate(self):
    interval = genome.Interval('chr1', 100, 200)
    self.assertIsNot(interval, interval.pad(0, 0))

  def test_pad_unstranded(self):
    self.assertEqual(
        genome.Interval('chr1', 100, 200, '.').pad(20, 10, use_strand=False),
        genome.Interval('chr1', 80, 210, '.'),
    )
    self.assertEqual(
        genome.Interval('chr1', 100, 200, '+').pad(20, 10, use_strand=False),
        genome.Interval('chr1', 80, 210, '+'),
    )
    self.assertEqual(
        genome.Interval('chr1', 100, 200, '+').pad(20, 10, use_strand=False),
        genome.Interval('chr1', 80, 210, '+'),
    )

  def test_pad_stranded(self):
    self.assertEqual(
        genome.Interval('chr1', 100, 200, '.').pad(20, 10, use_strand=True),
        genome.Interval('chr1', 80, 210, '.'),
    )
    self.assertEqual(
        genome.Interval('chr1', 100, 200, '+').pad(20, 10, use_strand=True),
        genome.Interval('chr1', 80, 210, '+'),
    )
    # Negative strand interval padding behaves differently when use_strand=True
    self.assertEqual(
        genome.Interval('chr1', 100, 200, '-').pad(20, 10, use_strand=True),
        genome.Interval('chr1', 90, 220, '-'),
    )

  def test_boundary_shift_positive_strand(self):
    interval = genome.Interval('chr1', 10, 20, strand='+')
    resized = interval.boundary_shift(1, -2)
    self.assertEqual(resized.start, 11)
    self.assertEqual(resized.end, 18)

  def test_boundary_shift_negative_strand(self):
    interval = genome.Interval('chr1', 10, 20, strand='-')
    resized = interval.boundary_shift(1, -2)
    self.assertEqual(resized.start, 12)
    self.assertEqual(resized.end, 19)

  def test_resize_positive_strand_odd_width(self):
    interval = genome.Interval('chr1', 10, 20, strand='+')
    i2 = interval.resize(11)
    self.assertEqual(i2.start, 9)
    self.assertEqual(i2.end, 20)

  def test_resize_positive_strand_even_width(self):
    interval = genome.Interval('chr1', 10, 20, strand='+')
    i2 = interval.resize(12)
    self.assertEqual(i2.start, 9)
    self.assertEqual(i2.end, 21)

  def test_resize_negative_strand_odd_width(self):
    interval = genome.Interval('chr1', 10, 20, strand='-')
    i2 = interval.resize(11)
    self.assertEqual(i2.start, 10)
    self.assertEqual(i2.end, 21)

  def test_resize_negative_strand_even_width(self):
    interval = genome.Interval('chr1', 10, 20, strand='-')
    i2 = interval.resize(12)
    self.assertEqual(i2.start, 9)
    self.assertEqual(i2.end, 21)

    i2 = interval.resize(9)
    self.assertEqual(i2.start, 11)
    self.assertEqual(i2.end, 20)

  @parameterized.parameters(('+', '-'), ('-', '+'))
  def test_swap_strand(self, before, after):
    interval = genome.Interval('chr1', 10, 20, strand=before)
    self.assertEqual(interval.swap_strand().strand, after)
    self.assertEqual(interval.strand, before)

  @parameterized.parameters(('+', '.'), ('-', '.'), ('.', '.'))
  def test_unstranded(self, before, after):
    interval = genome.Interval('chr1', 10, 20, strand=before)
    self.assertEqual(interval.as_unstranded().strand, after)
    self.assertEqual(interval.strand, before)

  def test_swap_strand_unstranded(self):
    interval = genome.Interval('chr1', 10, 20, strand='.')
    with self.assertRaises(ValueError):
      interval.swap_strand()

  @parameterized.parameters(
      (0, True), (0, False), (1, True), (1, False), (2, True), (2, False)
  )
  def test_center_even(self, shift, use_strand):
    """Checks even-width intervals."""
    interval = genome.Interval('chr1', 10 + shift, 20 + shift, strand='-')
    self.assertEqual(interval.center(use_strand=use_strand), 15 + shift)

  @parameterized.parameters(0, 1, 2)
  def test_center_odd_positive_strand(self, shift):
    """Checks odd-width intervals."""
    interval = genome.Interval('chr1', 9 + shift, 20 + shift, strand='+')
    self.assertEqual(interval.center(use_strand=False), 15 + shift)
    self.assertEqual(interval.center(use_strand=True), 15 + shift)

  @parameterized.parameters(0, 1, 2)
  def test_center_odd_negative_strand(self, shift):
    interval = genome.Interval('chr1', 9 + shift, 20 + shift, strand='-')
    self.assertEqual(interval.center(use_strand=False), 15 + shift)
    self.assertEqual(interval.center(use_strand=True), 14 + shift)

  @parameterized.parameters(
      (10, 13, '+'),  # even -> odd
      (10, 14, '+'),  # even -> even
      (11, 13, '+'),  # odd -> odd
      (11, 14, '+'),  # odd -> even
      (10, 13, '-'),  # negative strand
      (10, 14, '-'),
      (11, 13, '-'),
      (11, 14, '-'),
  )
  def test_center_constant_at_resize(self, start_width, resize_width, strand):
    """Tests that the center remains the same if we resize."""
    interval = genome.Interval('chr1', 10, 10 + start_width, strand=strand)
    self.assertEqual(interval.resize(resize_width).center(), interval.center())

  def test_intersect_overlaps(self):
    a = genome.Interval('chr1', 1, 5)
    b = genome.Interval('chr1', 3, 7)
    self.assertEqual(a.intersect(b), genome.Interval('chr1', 3, 5))

  def test_intersect_no_overlap(self):
    a = genome.Interval('chr1', 1, 5)
    b = genome.Interval('chr1', 7, 10)
    self.assertIsNone(a.intersect(b))

  def test_coverage(self):
    a = genome.Interval('chr1', 1, 5)
    b = genome.Interval('chr1', 3, 7)
    c = genome.Interval('chr1', 4, 7)
    d = genome.Interval('chr1', 1, 2)
    np.testing.assert_array_equal(a.coverage([b, c, d]), [1, 0, 1, 2])
    np.testing.assert_array_equal(a.coverage([b, c, d], bin_size=2), [1, 3])
    np.testing.assert_array_equal(a.coverage([b, c, d], bin_size=4), [4])

  def test_coverage_raises(self):
    a = genome.Interval('chr1', 1, 5)
    b = genome.Interval('chr1', 3, 7)
    c = genome.Interval('chr1', 4, 7)
    d = genome.Interval('chr1', 1, 2)
    with self.assertRaises(ValueError):
      a.coverage([b, c, d], bin_size=0)
    with self.assertRaises(ValueError):
      a.coverage([b, c, d], bin_size=-1)
    with self.assertRaises(ValueError):
      a.coverage([b, c, d], bin_size=3)

  def test_overlap_ranges(self):
    a = genome.Interval('chr1', 1, 5)
    b = genome.Interval('chr1', 3, 7)
    c = genome.Interval('chr1', 2, 4)
    d = genome.Interval('chr1', 0, 2)
    e = genome.Interval('chr1', 5, 10)
    f = genome.Interval('chr2', 1, 5)
    np.testing.assert_array_equal(
        a.overlap_ranges([b, c, d, e, f]),
        np.array(
            [
                (2, 4),
                (1, 3),
                (0, 1),
            ],
            dtype=np.int32,
        ),
    )
    np.testing.assert_array_equal(
        a.overlap_ranges([f]), np.empty((0, 2), dtype=np.int32)
    )

  def test_binary_mask(self):
    a = genome.Interval('chr1', 1, 5)
    b = genome.Interval('chr1', 3, 7)
    c = genome.Interval('chr1', 4, 7)
    d = genome.Interval('chr1', 1, 2)
    np.testing.assert_array_equal(
        a.binary_mask([b, c, d]), [True, False, True, True]
    )
    np.testing.assert_array_equal(
        a.binary_mask([b, c, d], bin_size=2), [True, True]
    )

  def test_binary_mask2(self):
    interval = genome.Interval('chr1', 0, 6)
    intervals = [
        genome.Interval('chr1', 1, 2),
        genome.Interval('chr1', 2, 3),
        genome.Interval('chr1', 3, 4),
    ]
    mask = interval.binary_mask(intervals, bin_size=1)
    self.assertSequenceEqual(
        [False, True, True, True, False, False], list(mask)
    )

    mask = interval.binary_mask(intervals, bin_size=2)
    self.assertSequenceEqual([True, True, False], list(mask))

  def test_coverage_stranded(self):
    a = genome.Interval('chr1', 1, 5)
    a_reversed = genome.Interval('chr1', 1, 5, strand='-')
    b = genome.Interval('chr1', 3, 7)
    c = genome.Interval('chr1', 4, 7)
    d = genome.Interval('chr1', 1, 2, strand='-')
    np.testing.assert_array_equal(
        a.coverage_stranded([b, c, d]), [[0, 1], [0, 0], [1, 0], [2, 0]]
    )
    np.testing.assert_array_equal(
        a_reversed.coverage_stranded([b, c, d]),
        [[0, 2], [0, 1], [0, 0], [1, 0]],
    )

  def test_binary_mask_stranded(self):
    interval = genome.Interval('chr1', 0, 6)
    intervals = [
        genome.Interval('chr1', 1, 2),
        genome.Interval('chr1', 2, 3),
        genome.Interval('chr1', 3, 4),
    ]

    mask = interval.binary_mask_stranded(intervals, bin_size=2)
    np.testing.assert_array_equal(
        [[True, False], [True, False], [False, False]], mask
    )

  def test_binary_mask_stranded2(self):
    interval = genome.Interval('chr1', 0, 6, strand='-')
    intervals = [
        genome.Interval('chr1', 1, 2, strand='+'),
        genome.Interval('chr1', 2, 3, strand='+'),
        genome.Interval('chr1', 3, 4, strand='-'),
    ]

    mask = interval.binary_mask_stranded(intervals, bin_size=2)
    np.testing.assert_array_equal(
        [[False, False], [True, True], [False, True]], mask
    )


class VariantTest(parameterized.TestCase):

  def test_proto_roundtrip(self):
    variant = genome.Variant('chr1', 10, 'C', 'T')
    proto = variant.to_proto()
    round_trip = genome.Variant.from_proto(proto)
    self.assertEqual(variant, round_trip)

  def test_init_valid(self):
    genome.Variant('chr1', 10, 'C', 'T')
    genome.Variant('chr1', 10, 'C', 'T', name='test')
    genome.Variant(
        'chr1',
        10,
        'C',
        'T',
        name='test',
        info=dict(a=1, b='b', c=[1, 2], d=dict(e=1)),
    )

  @parameterized.parameters(
      ('chr1', -20, 'C', 'T', 'Position has to be >=1.'),
      ('chr1', 1, 'Z', 'T', 'Invalid reference bases'),
      ('chr1', 1, 'A', 'Z', 'Invalid alternate bases'),
  )
  def test_init_raises_value_error(
      self, chromosome, position, ref_bases, alt_bases, expected_error
  ):
    with self.assertRaisesRegex(ValueError, expected_error):
      genome.Variant(chromosome, position, ref_bases, alt_bases)

  def test_attributes(self):
    v = genome.Variant('chr1', 10, 'C', 'T')
    self.assertEqual(v.start, 9)
    self.assertEqual(v.chromosome, 'chr1')
    self.assertEqual(v.position, 10)
    self.assertEqual(v.reference_bases, 'C')
    self.assertEqual(v.alternate_bases, 'T')

  def test_attributes_info_assignment(self):
    v = genome.Variant('chr1', 10, 'C', 'T')
    self.assertIsInstance(v.info, dict)
    v.info['test'] = 10
    self.assertEqual(v.info['test'], 10)

  def test_copy(self):
    # Make sure the original got unchanged.
    v = genome.Variant('chr1', 10, 'C', 'T', info=dict(test=10))
    v2 = v.copy()
    v.info['test'] = 20
    self.assertEqual(v2.info['test'], 10)
    self.assertEqual(v.info['test'], 20)

  @parameterized.parameters(
      ('chr22:1024:A>C', genome.VariantFormat.DEFAULT),
      ('chr22_1024_A_C_b38', genome.VariantFormat.GTEX),
      ('22_1024_A_C', genome.VariantFormat.OPEN_TARGETS),
      ('22:1024:A:C', genome.VariantFormat.OPEN_TARGETS_BIGQUERY),
      ('22-1024-A-C', genome.VariantFormat.GNOMAD),
  )
  def test_from_str_single_base(self, string, variant_format):
    expected_variant = genome.Variant(
        chromosome='chr22',
        position=1024,
        reference_bases='A',
        alternate_bases='C',
    )
    actual_variant = genome.Variant.from_str(
        string, variant_format=variant_format
    )
    self.assertEqual(expected_variant.chromosome, actual_variant.chromosome)
    self.assertEqual(expected_variant.position, actual_variant.position)
    self.assertEqual(
        expected_variant.reference_bases, actual_variant.reference_bases
    )
    self.assertEqual(
        expected_variant.alternate_bases, actual_variant.alternate_bases
    )

  @parameterized.parameters(
      ('chrX:1024:AGTC>C', genome.VariantFormat.DEFAULT),
      ('chrX_1024_AGTC_C', genome.VariantFormat.GTEX),
      ('X_1024_AGTC_C', genome.VariantFormat.OPEN_TARGETS),
      ('X:1024:AGTC:C', genome.VariantFormat.OPEN_TARGETS_BIGQUERY),
      ('X-1024-AGTC-C', genome.VariantFormat.GNOMAD),
  )
  def test_from_str_multiple_bases(self, string, variant_format):
    expected_variant = genome.Variant(
        chromosome='chrX',
        position=1024,
        reference_bases='AGTC',
        alternate_bases='C',
    )
    actual_variant = genome.Variant.from_str(
        string, variant_format=variant_format
    )
    self.assertEqual(expected_variant.chromosome, actual_variant.chromosome)
    self.assertEqual(expected_variant.position, actual_variant.position)
    self.assertEqual(
        expected_variant.reference_bases, actual_variant.reference_bases
    )
    self.assertEqual(
        expected_variant.alternate_bases, actual_variant.alternate_bases
    )

  @parameterized.parameters(
      ('chrM:1024:AGTC>', genome.VariantFormat.DEFAULT),
      ('chrM_1024_AGTC__hg38', genome.VariantFormat.GTEX),
      ('M_1024_AGTC_', genome.VariantFormat.OPEN_TARGETS),
      ('M:1024:AGTC:', genome.VariantFormat.OPEN_TARGETS_BIGQUERY),
      ('M-1024-AGTC-', genome.VariantFormat.GNOMAD),
  )
  def test_from_str_deletion(self, string, variant_format):
    expected_variant = genome.Variant(
        chromosome='chrM',
        position=1024,
        reference_bases='AGTC',
        alternate_bases='',
    )
    actual_variant = genome.Variant.from_str(
        string, variant_format=variant_format
    )
    self.assertEqual(expected_variant.chromosome, actual_variant.chromosome)
    self.assertEqual(expected_variant.position, actual_variant.position)
    self.assertEqual(
        expected_variant.reference_bases, actual_variant.reference_bases
    )
    self.assertEqual(
        expected_variant.alternate_bases, actual_variant.alternate_bases
    )

  @parameterized.parameters(
      ('chrM:1024:>AGTC', genome.VariantFormat.DEFAULT),
      ('chrM_1024__AGTC_hg38', genome.VariantFormat.GTEX),
      ('M_1024__AGTC', genome.VariantFormat.OPEN_TARGETS),
      ('M:1024::AGTC', genome.VariantFormat.OPEN_TARGETS_BIGQUERY),
      ('M-1024--AGTC', genome.VariantFormat.GNOMAD),
  )
  def test_from_str_insertion(self, string, variant_format):
    expected_variant = genome.Variant(
        chromosome='chrM',
        position=1024,
        reference_bases='',
        alternate_bases='AGTC',
    )
    actual_variant = genome.Variant.from_str(
        string, variant_format=variant_format
    )
    self.assertEqual(expected_variant.chromosome, actual_variant.chromosome)
    self.assertEqual(expected_variant.position, actual_variant.position)
    self.assertEqual(
        expected_variant.reference_bases, actual_variant.reference_bases
    )
    self.assertEqual(
        expected_variant.alternate_bases, actual_variant.alternate_bases
    )

  @parameterized.parameters(
      # Invalid chromosomes.
      ('chrR:1024:A>C', genome.VariantFormat.DEFAULT),
      ('chrR_1024_A_C', genome.VariantFormat.GTEX),
      ('R_1024_A_C', genome.VariantFormat.OPEN_TARGETS),
      ('R:1024:A:C', genome.VariantFormat.OPEN_TARGETS_BIGQUERY),
      ('R-1024-A-C', genome.VariantFormat.GNOMAD),
      # Missing chromosomes.
      (':1024:A>C', genome.VariantFormat.DEFAULT),
      ('_1024_A_C', genome.VariantFormat.GTEX),
      ('_1024_A_C', genome.VariantFormat.OPEN_TARGETS),
      (':1024:A:C', genome.VariantFormat.OPEN_TARGETS_BIGQUERY),
      ('-1024-A-C', genome.VariantFormat.GNOMAD),
      # Missing positions.
      ('chr22::A>C', genome.VariantFormat.DEFAULT),
      ('chr22__A_C', genome.VariantFormat.GTEX),
      ('22__A_C', genome.VariantFormat.OPEN_TARGETS),
      ('22::A:C', genome.VariantFormat.OPEN_TARGETS_BIGQUERY),
      ('22--A-C', genome.VariantFormat.GNOMAD),
      # Missing chr prefixes.
      ('22:1024:A>C', genome.VariantFormat.DEFAULT),
      ('22_1024_A_C', genome.VariantFormat.GTEX),
      # Unexpected chr prefixes.
      ('chr22_1024_A_C', genome.VariantFormat.OPEN_TARGETS),
      ('chr22:1024:A:C', genome.VariantFormat.OPEN_TARGETS_BIGQUERY),
      ('chr22-1024-A-C', genome.VariantFormat.GNOMAD),
      # Unexpected builds.
      ('chr22:1024:A>C_b38', genome.VariantFormat.DEFAULT),
      ('22_1024_A_C_b', genome.VariantFormat.OPEN_TARGETS),
      ('22:1024:A:C_hg38', genome.VariantFormat.OPEN_TARGETS_BIGQUERY),
      # Mismatched formats.
      ('22-1024-A-C_b38', genome.VariantFormat.GNOMAD),
      ('chr22_1024_A_C', genome.VariantFormat.DEFAULT),
      ('chr22:1024:A>C', genome.VariantFormat.GTEX),
      ('chr22:1024:A>C', genome.VariantFormat.OPEN_TARGETS),
      ('chr22:1024:A>C', genome.VariantFormat.OPEN_TARGETS_BIGQUERY),
      ('chr22:1024:A>C', genome.VariantFormat.GNOMAD),
  )
  def test_from_str_invalid(self, string, variant_format):
    with self.assertRaisesRegex(
        ValueError, f'Invalid format for variant string: {string}'
    ):
      genome.Variant.from_str(string, variant_format=variant_format)

  def test_from_to_string(self):
    v = genome.Variant('chr1', 10, 'C', 'T')
    self.assertIsInstance(str(v), str)
    self.assertEqual(v, genome.Variant.from_str(str(v)))

  def test_from_to_dict(self):
    v = genome.Variant('chr1', 10, 'C', 'T')
    d = v.to_dict()
    self.assertIsInstance(d, dict)
    self.assertEqual(v, genome.Variant.from_dict(d))

  def test_split_alternate_longer(self):
    # Case #1: alternate_bases longer than reference_bases.
    v = genome.Variant('chr1', 3, 'AC', 'TGTC')
    # Anchor before the variant.
    self.assertEqual(
        v.split(2), (None, genome.Variant('chr1', 3, 'AC', 'TGTC'))
    )
    # Anchor at the variant (between A and C in the reference).
    # Note: start=2, end=4.
    self.assertEqual(
        v.split(3),
        (
            genome.Variant('chr1', 3, 'A', 'T'),
            genome.Variant('chr1', 4, 'C', 'GTC'),
        ),
    )
    # Anchor after the variant.
    for split in [4, 5, 6]:
      self.assertEqual(
          v.split(split), (genome.Variant('chr1', 3, 'AC', 'TGTC'), None)
      )

  def test_split_reference_longer(self):
    # Case #2: reference_bases longer than alternate_bases.
    v = genome.Variant('chr1', 3, 'TGTC', 'AC')
    # Anchor before the variant.
    self.assertEqual(
        v.split(2), (None, genome.Variant('chr1', 3, 'TGTC', 'AC'))
    )
    # Anchor at the variant.
    self.assertEqual(
        v.split(3),
        (
            genome.Variant('chr1', 3, 'T', 'A'),
            genome.Variant('chr1', 4, 'GTC', 'C'),
        ),
    )
    self.assertEqual(
        v.split(4),
        (
            genome.Variant('chr1', 3, 'TG', 'AC'),
            genome.Variant('chr1', 5, 'TC', ''),
        ),
    )
    self.assertEqual(
        v.split(5),
        (
            genome.Variant('chr1', 3, 'TGT', 'AC'),
            genome.Variant('chr1', 6, 'C', ''),
        ),
    )
    # Anchor after the variant.
    self.assertEqual(
        v.split(6), (genome.Variant('chr1', 3, 'TGTC', 'AC'), None)
    )

  @parameterized.parameters(
      ('A', 'C', True),
      ('AC', 'A', False),
      ('A', 'AC', False),
      ('AC', 'GT', False),
      ('', 'A', False),
  )
  def test_is_snv(self, ref, alt, expected):
    v = genome.Variant('chr1', 10, ref, alt)
    self.assertEqual(v.is_snv, expected)

  @parameterized.parameters(
      ('A', 'C', False),
      ('AC', 'A', True),
      ('A', 'AC', False),
      ('AC', 'GT', False),
      ('', 'A', False),
      ('A', '', True),
  )
  def test_is_deletion(self, ref, alt, expected):
    v = genome.Variant('chr1', 10, ref, alt)
    self.assertEqual(v.is_deletion, expected)

  @parameterized.parameters(
      ('A', 'C', False),
      ('AC', 'A', False),
      ('A', 'AC', True),
      ('AC', 'GT', False),
      ('', 'A', True),
  )
  def test_is_insertion(self, ref, alt, expected):
    v = genome.Variant('chr1', 10, ref, alt)
    self.assertEqual(v.is_insertion, expected)

  @parameterized.parameters(
      ('chr1:1:AAA>GGG', 'TTTNNNN', 'chr1:1:AAA>GGG'),
      ('chr1:1:ATCG>ATCC', 'ATCGNNN', 'chr1:4:G>C'),
      ('chr1:1:ATCG>ATC', 'ATCG', 'chr1:3:CG>C'),
      ('chr1:1:ACTG>GGTG', 'NNNN', 'chr1:1:ACTG>GGTG'),
      ('chr1:1:C>AAA', 'CCC', 'chr1:1:C>AAA'),
      ('chr1:5:C>', 'AAAAA', 'chr1:4:AC>A'),
      ('chr1:2:A>A', 'AAAAAA', 'chr1:2:A>A'),
      ('chr1:1:GAT>GT', 'GAT', 'chr1:1:GA>G'),
      ('chr1:1:GAT>GT', 'GAN', 'chr1:2:AT>T'),
      # Examples from paper.
      ('chr1:6:CAC>C', 'GGGCACACACAGGG', 'chr1:3:GCA>G'),
      ('chr2:3:GCACA>GCA', 'GGGCACACACAGGG', 'chr2:3:GCA>G'),
      ('chr3:2:GGCA>GG', 'GGGCACACACAGGG', 'chr3:3:GCA>G'),
      ('chr4:3:GCA>G', 'GGGCACACACAGGG', 'chr4:3:GCA>G'),
  )
  def test_normalize_variant(
      self, variant: str, reference_sequence: str, expected: str
  ):
    mock_extractor = mock.create_autospec(genome._FastaExtractorType)  # pylint: disable=protected-access
    variant = genome.Variant.from_str(variant)

    def _extract(interval):
      return reference_sequence[interval.start : interval.end]

    mock_extractor.extract.side_effect = _extract

    result = genome.normalize_variant(variant, mock_extractor)
    self.assertEqual(str(result), expected)


class JunctionTest(parameterized.TestCase):

  def test_init_valid(self):
    genome.Junction(chromosome='chr1', start=10, end=10, strand='+')
    genome.Junction(chromosome='chr1', start=10, end=20, strand='-', k=1)

  def test_unstranded_junction_raises_value_error(self):
    with self.assertRaises(ValueError):
      genome.Junction(chromosome='chr1', start=10, end=10)

  def test_invalid_start_end_raises_value_error(self):
    with self.assertRaises(ValueError):
      genome.Junction(chromosome='chr1', start=10, end=1, strand='+')

  def test_attributes(self):
    junction = genome.Junction(
        'chr1',
        10,
        20,
        strand='-',
        info=dict(a=1, b='b', c=[1, 2], d=dict(e=1)),
        k=1,
    )
    self.assertEqual(junction.start, 10)
    self.assertEqual(junction.end, 20)
    self.assertEqual(junction.chromosome, 'chr1')
    self.assertEqual(junction.name, '')
    self.assertIsInstance(junction.info, dict)
    self.assertEqual(junction.info['a'], 1)
    self.assertTrue(junction.negative_strand)
    self.assertEqual(junction.width, 10)
    self.assertEqual(junction.k, 1)

  def test_copy(self):
    junction = genome.Junction('chr1', 10, 20, strand='-', k=1)
    junction_copy = junction.copy()
    junction_copy.info['test'] = 10
    self.assertEqual(junction_copy.info['test'], 10)
    self.assertEmpty(junction.info)
    self.assertEqual(junction.k, junction_copy.k)

  def test_acceptor_donor(self):
    junction = genome.Junction('chr1', 10, 20, strand='-')
    self.assertEqual(junction.acceptor, 10)
    self.assertEqual(junction.donor, 20)

    junction = genome.Junction('chr1', 10, 20, strand='+')
    self.assertEqual(junction.acceptor, 20)
    self.assertEqual(junction.donor, 10)

  def test_dinucleotide_region(self):
    junction = genome.Junction('chr1', 10, 20, strand='-')
    self.assertEqual(
        junction.dinucleotide_region(),
        (
            genome.Interval('chr1', 10, 12, '-'),
            genome.Interval('chr1', 18, 20, '-'),
        ),
    )

  def test_acceptor_donor_region(self):
    junction = genome.Junction('chr1', 10, 20, strand='-')
    self.assertEqual(
        junction.acceptor_region(),
        genome.Interval('chr1', -240, 260, '-'),
    )
    self.assertEqual(
        junction.donor_region(),
        genome.Interval('chr1', -230, 270, '-'),
    )


class IntervalIntersectionTest(parameterized.TestCase):

  def _parse_intervals(self, intervals):
    if not intervals:
      return []
    return [genome.Interval.from_str(i) for i in intervals.split(',')]

  @parameterized.named_parameters(
      ('no_overlap_1', 'chr1:0-100', 'chr2:0-100', ''),
      ('no_overlap_2', 'chr1:0-100', 'chr1:200-300', ''),
      ('overlap_1', 'chr1:0-100', 'chr1:50-150', 'chr1:50-100'),
      (
          'overlap_2',
          'chr1:0-100,chr2:100-200',
          'chr2:50-150,chr1:50-150',
          'chr1:50-100,chr2:100-150',
      ),
      (
          'overlap_3',
          'chr1:0-100,chr1:150-200',
          'chr1:50-200',
          'chr1:50-100,chr1:150-200',
      ),
      (
          'point_intervals',
          'chr1:10-10,chr1:15-15,chr1:20-20',
          'chr1:5-15',
          'chr1:10-10,chr1:15-15',
      ),
      ('point_intervals_2', 'chr1:10-10,chr1:10-20', 'chr1:5-15', 'chr1:10-15'),
      ('point_intervals_3', 'chr1:0-10,chr1:10-10', 'chr1:5-15', 'chr1:5-10'),
      (
          'abutting_intervals',
          'chr1:10-20,chr1:20-30',
          'chr1:15-25',
          'chr1:15-25',
      ),
      ('abutting_intervals_2', 'chr1:0-100', 'chr1:100-200', 'chr1:100-100'),
  )
  def test_intersection(self, lhs, rhs, expected):
    with self.subTest('forward'):
      self.assertSameElements(
          self._parse_intervals(expected),
          list(
              genome.intersect_intervals(
                  self._parse_intervals(lhs), self._parse_intervals(rhs)
              )
          ),
      )
    with self.subTest('reverse'):
      self.assertSameElements(
          self._parse_intervals(expected),
          list(
              genome.intersect_intervals(
                  self._parse_intervals(rhs), self._parse_intervals(lhs)
              )
          ),
      )


class IntervalUnionTest(parameterized.TestCase):

  def _parse_intervals(self, intervals):
    if not intervals:
      return []
    return [genome.Interval.from_str(i) for i in intervals.split(',')]

  @parameterized.named_parameters(
      ('no_overlap_1', 'chr1:0-100', 'chr2:0-100', 'chr1:0-100,chr2:0-100'),
      ('no_overlap_2', 'chr1:0-100', 'chr1:200-300', 'chr1:0-100,chr1:200-300'),
      ('overlap_1', 'chr1:0-100', 'chr1:50-150', 'chr1:0-150'),
      (
          'overlap_2',
          'chr1:0-100,chr2:100-200',
          'chr2:50-150,chr1:50-150',
          'chr1:0-150,chr2:50-200',
      ),
      ('overlap_3', 'chr1:0-100,chr1:150-200', 'chr1:50-200', 'chr1:0-200'),
      (
          'point_intervals',
          'chr1:10-10,chr1:15-15,chr1:20-20',
          'chr1:5-15',
          'chr1:5-15,chr1:20-20',
      ),
      ('point_intervals_2', 'chr1:10-10,chr1:10-20', 'chr1:5-15', 'chr1:5-20'),
      ('point_intervals_3', 'chr1:0-10,chr1:10-10', 'chr1:5-15', 'chr1:0-15'),
      (
          'abutting_intervals',
          'chr1:10-20,chr1:20-30',
          'chr1:15-25',
          'chr1:10-30',
      ),
      ('abutting_intervals_2', 'chr1:0-100', 'chr1:100-200', 'chr1:0-200'),
  )
  def test_union(self, lhs, rhs, expected):
    with self.subTest('forward'):
      self.assertSameElements(
          self._parse_intervals(expected),
          list(
              genome.union_intervals(
                  self._parse_intervals(lhs), self._parse_intervals(rhs)
              )
          ),
      )
    with self.subTest('reverse'):
      self.assertSameElements(
          self._parse_intervals(expected),
          list(
              genome.union_intervals(
                  self._parse_intervals(rhs), self._parse_intervals(lhs)
              )
          ),
      )


class MergeIntervalsTest(parameterized.TestCase):

  def test_empty_list(self):
    self.assertEmpty(genome.merge_overlapping_intervals([]))

  def test_one_interval(self):
    intervals = [genome.Interval('chr1', 1, 100)]
    self.assertListEqual(
        genome.merge_overlapping_intervals(intervals), intervals
    )

  def test_no_overlap(self):
    intervals = [
        genome.Interval('chr1', 1, 100),
        genome.Interval('chr1', 101, 200),
    ]
    self.assertListEqual(
        genome.merge_overlapping_intervals(intervals), intervals
    )

  def test_interval_within(self):
    intervals = [
        genome.Interval('chr1', 1, 100),
        genome.Interval('chr1', 10, 100),
    ]
    self.assertListEqual(
        genome.merge_overlapping_intervals(intervals), [intervals[0]]
    )

  def test_single_overlap(self):
    intervals = [
        genome.Interval('chr1', 1, 100),
        genome.Interval('chr1', 90, 200),
        genome.Interval('chr1', 300, 400),
    ]
    self.assertListEqual(
        genome.merge_overlapping_intervals(intervals),
        [genome.Interval('chr1', 1, 200), intervals[-1]],
    )

  def test_multiple_overlap(self):
    intervals = [
        genome.Interval('chr1', 1, 100),
        genome.Interval('chr1', 100, 200),
        genome.Interval('chr1', 200, 300),
        genome.Interval('chr1', 500, 700),
        genome.Interval('chr1', 700, 800),
        genome.Interval('chr1', 750, 801),
    ]
    self.assertListEqual(
        genome.merge_overlapping_intervals(intervals),
        [genome.Interval('chr1', 1, 300), genome.Interval('chr1', 500, 801)],
    )

  def test_multiple_overlap_unsorted(self):
    intervals = [
        genome.Interval('chr1', 1, 100),
        genome.Interval('chr1', 200, 300),
        genome.Interval('chr1', 700, 800),
        genome.Interval('chr1', 750, 801),
        genome.Interval('chr1', 500, 700),
        genome.Interval('chr1', 100, 200),
    ]
    self.assertListEqual(
        genome.merge_overlapping_intervals(intervals),
        [genome.Interval('chr1', 1, 300), genome.Interval('chr1', 500, 801)],
    )

  def test_info_field(self):
    original_intervals = [
        genome.Interval(
            'chr1', 1, 100, info={'transcript_type': 'protein_coding'}
        ),
        genome.Interval('chr1', 101, 200),
    ]
    merged_intervals = genome.merge_overlapping_intervals(original_intervals)
    self.assertListEqual(merged_intervals, original_intervals)
    for interval in merged_intervals:
      self.assertEmpty(interval.info)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plotting functions."""

from collections.abc import Callable, Mapping, Sequence
import math
from typing import Any

from absl import logging
from alphagenome.data import genome
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


def seqlogo(
    letter_heights: np.ndarray,
    alphabet: str = 'ACGT',
    one_based: bool = True,
    start: int = 0,
    ax: mpl.axis.Axis | None = None,
    letter_colors: str = 'default',
):
  """Sequence logo plot.

  Args:
    letter_heights: 2D array with shape (sequence length, len(alphabet)) shape.
      Positive values indicate that a logo element extends above the horizontal
      axis, and negative values indicate that the element extends below.
    alphabet: Alphabet corresponding to the channel axis of `letter_heights`.
    one_based: If True, plot letters at 1-based coordinates: first letter will
      be centered at 1. If False, the first letter will be between 0 and 1.
    start: Interval start.
    ax: matplotlib axis to add figures to.
    letter_colors: Name of the color scheme to use for coloring each letter.
      Must be one of 'default', a scheme consistent with weblogo
      (https://github.com/gecrooks/weblogo/blob/master/weblogo/logo.py#L136) or
      'factorbook', a scheme consistent with Factorbook
      (https://www.factorbook.org/).
  """
  ax = ax or plt.gca()

  if letter_heights.ndim != 2:
    raise ValueError(
        'Expecting a 2D matrix of shape (sequence_length, len(alphabed))'
    )
  if letter_heights.shape[1] != len(alphabet):
    raise ValueError('Last axis needs to match len(alphabet)')

  for x_offset, heights in enumerate(letter_heights):
    last_positive_y = 0.0
    last_negative_y = 0.0
    for height, letter in sorted(zip(heights, alphabet)):
      # Start with lowest height and keep track of the last y coordinate.
      if height > 0:
        y_position = last_positive_y
        last_positive_y += height
      else:
        y_position = last_negative_y
        last_negative_y += height
      _add_letter_to_axis(
          ax,
          letter,
          x=start + x_offset + 0.5 + 0.5 * one_based,
          y=y_position,
          height=height,
          letter_colors=letter_colors,
      )

  # Make sure all letters are displayed.
  ax.autoscale_view()  # pytype: disable=attribute-error


def _add_letter_to_axis(ax, letter, x, y, height=1, letter_colors='default'):
  """Add a letter with unit width and stretched by height to the axis."""
  globscale = 1.35
  letter_width_colors = dict(
      # Consistent with `classic` scheme from
      # https://github.com/gecrooks/weblogo/blob/master/weblogo/logo.py#L136
      default={
          'A': (0.35, 'darkgreen'),
          'C': (0.366, 'blue'),
          'G': (0.384, 'orange'),
          'T': (0.305, 'red'),
      },
      # Consistent with factorbook scheme from https://www.factorbook.org/
      factorbook={
          'A': (0.305, 'red'),
          'C': (0.366, 'blue'),
          'G': (0.384, 'orange'),
          'T': (0.35, 'darkgreen'),
      },
  )
  try:
    letter_width_color = letter_width_colors[letter_colors]
  except KeyError as e:
    raise ValueError(
        f'Letter_colors must be one of {letter_width_colors.keys()}.'
    ) from e

  font = mpl.font_manager.FontProperties(weight='bold')
  letter_width, letter_color = letter_width_color[letter]

  letter_transform = (
      mpl.transforms.Affine2D()
      .scale(globscale, height * globscale)
      .translate(x, y)
  )
  p = mpl.patches.PathPatch(
      mpl.text.TextPath((-letter_width, 0), letter, size=1, prop=font),
      lw=0,
      fc=letter_color,
      transform=letter_transform
      + ax.transData,  # pytype: disable=attribute-error
  )
  ax.add_patch(p)  # pytype: disable=attribute-error
  return p


def plot_contact_map(
    contact_map: pd.DataFrame,
    vmin: float | None = None,
    vmax: float | None = None,
    square: bool = True,
    cbar_shrink: float = 0.4,
    ax=None,
    **kwargs,
):
  """Visualize a contact map.

  A contact map is a 2D array of values, where each value represents the
  probability that two DNA bases are in contact.

  The array is symmetric, i.e., the (i, j) values and the (j, i) values of
  `contact_map` are equal.

  Args:
    contact_map: Contact map to visualize.
    vmin: Minimum value for the colorbar.
    vmax: Maximum value for the colorbar.
    square: If `True`, plots the contact map as a square. If `False`, plots the
      contact map as a rectangle with a shorter height than width.
    cbar_shrink: Fraction by which to multiply the size of the colorbar.
    ax: Matplotlib axis to add the heatmap to.
    **kwargs: Additional keyword arguments passed to `sns.heatmap`.
  """

  def tuple2string(interval_tuple):
    chromosome, start, end = interval_tuple
    return f'{chromosome}:{start:,}-{end:,}'

  if isinstance(contact_map, pd.DataFrame):
    interval_string = (
        tuple2string(contact_map.index[0])
        + ' - '
        + tuple2string(contact_map.index[-1])
    )
    contact_map = contact_map.values
  else:
    interval_string = ''

  if vmin is None:
    vmin = np.nanmin(contact_map[contact_map > 0])
  if vmax is None:
    vmax = np.nanmax(contact_map)
  fall_cmap = mpl.colors.LinearSegmentedColormap.from_list(
      'fall',
      colors=[
          'white',
          (245 / 256, 166 / 256, 35 / 256),
          (208 / 256, 2 / 256, 27 / 256),
          'black',
      ],
  )
  log_norm = mpl.colors.LogNorm(vmin=vmin, vmax=vmax)
  cbar_min = math.floor(math.log10(log_norm.vmin))
  cbar_max = 1 + math.ceil(math.log10(log_norm.vmax))
  cbar_ticks = [math.pow(10, i) for i in range(cbar_min, cbar_max)]

  if ax is None:
    ax = plt.gca()
  sns.heatmap(
      contact_map + 1e-10,
      cmap=fall_cmap,
      norm=log_norm,
      cbar_kws={'ticks': cbar_ticks, 'shrink': cbar_shrink},
      vmin=log_norm.vmin,
      vmax=log_norm.vmax,
      xticklabels=False,
      yticklabels=False,
      square=square,
      ax=ax,
      **kwargs,
  )

  ax.set_xlabel(interval_string)


def plot_track(
    arr: np.ndarray,
    ax: plt.Axes,
    x: np.ndarray | None = None,
    legend: bool = False,
    ylim: float | None = None,
    color: str | Sequence[str] | None = None,
    filled: bool = False,
) -> None:
  """Plot a single track on the axis after inferring track type from array.

  This function infers the type of track plot depending on channel
  dimensionality of `arr`:

    - If `arr` is a single-dimensional array, plot a single track line.
    - If `arr` is a 2D array, and interpreting the second dimension as the
      channel dimension:

      - Plot a sequence logo if the number of channels is 4 (which suggests one
        channel per DNA base).
      - Plot two overlapping line plots if the number of channels is 2 (which
        suggests one channel per DNA strand).

  Args:
    arr: array of values to plot.
    ax: matplotlib axis to use for plotting the values on.
    x: optional array of values to use for setting the x axis values. If absent,
      then sequential values starting from 1 to the length of the array arr will
      be used.
    legend: whether to draw a legend or not on the double stranded DNA plot.
    ylim: y axis limit.
    color: color(s) used for line plots.
    filled: whether to fill the space between the x axis and the line (or leave
      it as whitespace).

  Returns:
    None. Adds a plot to the provided axis.
  """
  sequence_length = len(arr)

  if x is None:
    x = np.arange(1, sequence_length + 1)  # One-based indexing.

  if arr.ndim == 1 or arr.shape[1] == 1:
    if color is not None:
      if isinstance(color, (list, tuple)):
        logging.warning(
            'A sequence of colors was passed to plot_track but the second '
            'array dimension is 1 suggesting single stranded DNA. Using the '
            'first entry %s as the line color.',
            color[0],
        )
        color = color[0]

    # In the case of a boolean array, we fill between values and the x axis and
    # simplify the y axis components.
    if arr.dtype == bool:
      ax.fill_between(x, arr, step='mid', color=color)
      ax.yaxis.set_ticks_position('none')
      ax.set_yticklabels([])
      ax.spines['left'].set_visible(False)
      ax.set_ylim([0, 1.5])

    else:
      if filled:
        ax.fill_between(x, np.ravel(arr), color=color)
      ax.plot(x, np.ravel(arr), color=color)

  elif arr.shape[1] == 4:
    seqlogo(arr, ax=ax)

  elif arr.shape[1] == 2:
    if color is not None:
      if not isinstance(color, Sequence) or len(color) != 2:
        raise ValueError(
            'Must pass sequence of 2 colors if plotting 2 dimensional array.'
        )
      color_1, color_2 = color
    else:
      color_1, color_2 = None, None
    ax.plot(x, arr[:, 0], label='pos', color=color_1)
    ax.plot(x, arr[:, 1], label='neg', color=color_2)
    if legend:
      ax.legend()
  else:
    raise ValueError(
        f'Do not know how to plot array with shape[1] != {arr.shape[1]}. '
        'Valid values are: 1, 2, or 4.'
    )
  if ylim is not None:
    ax.set_ylim(ylim)


def plot_tracks(
    tracks: Mapping[str, Any],
    x: np.ndarray | None = None,
    title: str | None = None,
    legend: bool = False,
    fig_width: float = 20,
    fig_track_height: float | Mapping[str, float] = 1.5,
    ylim: str | Mapping[str, tuple[int, int]] | tuple[int, int] | None = 'auto',
    yticks_min_max_only: bool = False,
    ylab: bool = True,
    color: str | Mapping[str, str] | None = None,
    horizontal_ylab: bool = True,
    filled_tracks: Sequence[str] | None = None,
    despine: bool = True,
    despine_keep_bottom: bool = False,
    plot_track_fn: Callable[..., Any] = plot_track,
) -> mpl.figure.Figure:
  """Plot multiple tracks as subplots within one matplotlib figure.

  Custom plotting functions may be passed as `plot_track_fn`. Otherwise,
  `plot_track` will be used by default.

  Args:
    tracks: dictionary of tracks to plot. Example input: tracks = {'a': [1,2,3],
      'b': [4,5,4]}. The arrays in the dict must have the same length.
    x: optional array of x axis values. If absent, these will be inferred to be
      from 1 to the length of the arrays in tracks.
    title: Optional title to be added to the first subplot in the figure.
    legend: Whether to plot a legend or not.
    fig_width: Figure width.
    fig_track_height: Either a scalar specifying the total height of the figure,
      or a dictionary specifying the height of the subplot for each track.
    ylim: y axis limit. Must be either "same" (shared y limit across subplots)
      or "auto" (free scales, indicating separate limits per subplot).
    yticks_min_max_only: whether the only y axis ticks should be the min and max
      values.
    ylab: y axis label.
    color:  Either a string or a dict indicating a color per track.
    horizontal_ylab: Whether to display the y axis label horizontally (instead
      of the default vertical orientation).
    filled_tracks: List of track names that should have filled line plots drawn
      (instead of the default whitespace between the values and x axis).
    despine: Whether to remove top, right, and bottom spines.
    despine_keep_bottom: Whether to remove top and right spines, but keep the
      bottom spine.
    plot_track_fn: Optional custom function for plotting tracks. If absent,
      defaults to `plot_track`.

  Returns:
    Matplotlib figure.
  """
  # Set figure height (either total height, or per subplot if passed a dict).
  if isinstance(fig_track_height, dict):
    fig_height = sum(fig_track_height.values())
    gridspec_kw = {
        'height_ratios': [
            height / fig_height for height in fig_track_height.values()
        ]
    }
  else:
    fig_height = fig_track_height
    gridspec_kw = dict()

  # Generate figure axes.
  fig, axes = plt.subplots(
      nrows=len(tracks),
      ncols=1,
      figsize=(fig_width, fig_height),
      gridspec_kw=gridspec_kw,
      sharex=True,
  )
  if len(tracks) == 1:
    axes = [axes]

  # Set y axis limits.
  if ylim == 'same':
    ylim = (
        0,
        max(v.max() for v in tracks.values() if isinstance(v, np.ndarray)),
    )
  elif ylim == 'auto':
    ylim = None
  elif isinstance(ylim, str):
    raise ValueError('Only "same" and "auto" are valid strings for ylim.')

  # Iterate over tracks and plot one track per axis in the figure.
  for i, (ax, (track, arr)) in enumerate(zip(axes, tracks.items())):
    # Only set title in the first subplot.
    if i == 0 and title is not None:
      ax.set_title(title)

    # Only set x axis ticks in the final subplot.
    if i != len(tracks) - 1:
      ax.xaxis.set_ticks_position('none')

    track_ylim = ylim[track] if isinstance(ylim, dict) else ylim
    track_color = color[track] if isinstance(color, dict) else color

    # If the array is in fact a callable, then it is intended to be used as a
    # custom plotting function (most commonly, for plotting transcripts).
    if hasattr(arr, '__call__'):
      # TODO: b/377225766 - Allow custom plot functions as dictionary elements.
      arr(ax, ylim=ylim, color=color)
    else:
      filled_tracks = filled_tracks or []
      plot_track_fn(
          arr,
          ax=ax,
          x=x,
          legend=legend,
          ylim=track_ylim,
          color=track_color,
          filled=track in filled_tracks,
      )

    # Set y axis label.
    if ylab:
      if horizontal_ylab:
        ax.set_ylabel(
            track,
            rotation=0,
            multialignment='center',
            va='center',
            ha='right',
            labelpad=5,
        )
      else:
        ax.set_ylabel(track)

    # Only draw the y axis ticks on the min and max value locations.
    if yticks_min_max_only:
      if track_ylim is None:
        track_ylim = ax.get_yticks()[[0, -1]]
      # TODO: b/377226499 - Preventing negative values might be sub-optimal.
      minimum = max(track_ylim[0], 0)
      maximum = track_ylim[1]
      ax.spines['left'].set_bounds(minimum, maximum)
      ax.set_yticks([minimum, maximum])

    if despine:
      ax.spines['top'].set_visible(False)
      ax.spines['right'].set_visible(False)
      if i != len(tracks) - 1 and despine_keep_bottom:
        ax.spines['bottom'].set_visible(True)
      else:
        ax.spines['bottom'].set_visible(False)

  # Enable default tick locator for the final subplot.
  axes[-1].xaxis.set_major_locator(mpl.ticker.AutoLocator())

  # No height for whitespace between subplots.
  fig.subplots_adjust(hspace=0)

  return fig


def sashimi_plot(
    junctions: Sequence[genome.Junction],
    ax: plt.Axes,
    interval: genome.Interval | None = None,
    filter_threshold: float = 0.01,
    annotate_counts: bool = True,
    rng: np.random.Generator | None = None,
):
  """Plot splice junctions as a Sashimi plot.

  Sashimi plots (first described [here](https://arxiv.org/abs/1306.3466)
  visualize splice junctions from an RNA-seq experiment.

  According to the authors,
  * genomic reads are converted into read densities
  * junction reads are plotted as arcs whose width is determined by the number
  of reads aligned to the junction spanning the exons connected by the arc.

  Args:
    junctions: Splice junctions to plot.
    ax: Matplotlib axis to add the plot to.
    interval: Interval to plot, used to filter text annotations.
    filter_threshold: Junctions with number of reads below this threshold will
      be filtered out.
    annotate_counts: Whether to annotate the junctions with read counts.
    rng: Optional random number generator to use for jittering junction paths.
      If unset will use NumPy's default random number generator.
  """
  rng = rng or np.random.default_rng()
  total = np.sum([junction.k for junction in junctions])
  # Random jitter position to avoid overlap.
  jitters = rng.uniform(low=0.05, high=0.15, size=len(junctions))
  for junction, jt in zip(junctions, jitters):
    if junction.k < filter_threshold:
      continue
    k = junction.k / total
    verts = [
        (junction.start, 0.0),
        (junction.start, jt),
        (junction.end, jt),
        (junction.end, 0.0),
    ]

    path = mpl.path.Path(
        verts,
        [
            mpl.path.Path.MOVETO,
            mpl.path.Path.CURVE4,
            mpl.path.Path.CURVE4,
            mpl.path.Path.CURVE4,
        ],
    )
    patch = mpl.patches.PathPatch(path, facecolor='none', lw=min(k * 30, 5))
    ax.add_patch(patch)
    if annotate_counts:
      text_pos = junction.center()
      if interval is not None:
        if text_pos < interval.start or text_pos > interval.end:
          continue
      ax.text(text_pos, 0.8 * jt, junction.k, horizontalalignment='center')
  ax.set_ylim(0, 0.15)


def pad_track(track: np.ndarray, new_len: int, value: int = 0) -> np.ndarray:
  """Pad a track with `value` to the desired length.

  If new_len - len(track) is an even number, the same amount of padding is added
  to both sides.

  If new_len - len(track) is an odd number, the extra padded element will be
  added at the end of the array.

  Args:
    track: Track to pad.
    new_len: Desired length of the padded track.
    value: Value to use for padding.

  Returns:
    Padded track.
  """
  assert track.ndim == 2
  assert new_len >= len(track)

  out = np.empty((new_len, track.shape[1]))
  out[:] = value
  delta = new_len - len(track)
  i = delta // 2
  j = i + len(track)
  out[i:j] = track
  return out



# Copyright 2025 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for type annotations."""

import importlib.metadata
from typing import TypeVar

import jaxtyping
import typeguard

_T = TypeVar('_T')


def jaxtyped(fn: _T) -> _T:
  """Wrapper around jaxtyping.jaxtyped that uses typeguard iff typeguard < 3."""
  try:
    major, *_ = importlib.metadata.version('typeguard').split('.')
  except importlib.metadata.PackageNotFoundError:
    major = -1

  # Only use jaxtyping if typeguard is < 3. See
  # https://docs.kidger.site/jaxtyping/api/runtime-type-checking/#runtime-type-checking
  # for more details.
  if int(major) < 3:
    return jaxtyping.jaxtyped(fn, typechecker=typeguard.typechecked)
  else:
    return fn



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Splice junction data container.

`JunctionData` stores splice junction data for a given transcript or interval.
"""

from collections.abc import Sequence
import dataclasses
from typing import Any

from alphagenome import typing
from alphagenome.data import genome
from alphagenome.data import ontology
from jaxtyping import Float, Shaped  # pylint: disable=g-multiple-import, g-importing-member
import numpy as np
import pandas as pd

JunctionMetadata = pd.DataFrame


@typing.jaxtyped
@dataclasses.dataclass(frozen=True)
class JunctionData:
  """Container for storing splice junction data.

  Attributes:
    junctions: A numpy array representing the splice junctions.
    values: A numpy array of floats representing the values associated with each
      junction for each track.
    metadata: A pandas DataFrame containing metadata for each track.
    interval: An optional `Interval` object representing the genomic region
      containing the junctions.
    uns: An optional dictionary to store additional unstructured data.

  Raises:
    ValueError: If the number of tracks in `values` does not match the number
      of rows in `metadata`, or if `metadata` contains duplicate names.
  """

  junctions: Shaped[np.ndarray, 'num_junctions']
  values: Float[np.ndarray, 'num_junctions num_tracks']
  metadata: JunctionMetadata
  interval: genome.Interval | None = None
  uns: dict[str, Any] | None = None

  def __post_init__(self):
    """Validates the consistency of the data."""
    if self.values.shape[1] != len(self.metadata):
      raise ValueError(
          f'Number of tracks {self.values.shape[1]} and '
          f'metadata {len(self.metadata)} do not match.'
      )

    if self.metadata['name'].duplicated().any():
      raise ValueError('Metadata contain duplicated names.')

  def __len__(self):
    """Returns the number of junctions."""
    return len(self.junctions)

  @property
  def num_tracks(self) -> int:
    """Returns the number of tracks."""
    return len(self.metadata)

  @property
  def names(self) -> np.ndarray:
    """Returns a list of track names (not necessarily unique)."""
    return self.metadata['name'].values

  @property
  def strands(self) -> np.ndarray:
    """Returns a list of track strands."""
    return np.array([j.strand for j in self.junctions])

  @property
  def possible_strands(self) -> np.ndarray:
    """All possible strands."""
    return np.unique(self.strands)

  @property
  def ontology_terms(self) -> Sequence[ontology.OntologyTerm | None] | None:
    """Returns a list of ontology terms (if available)."""
    if 'ontology_curie' in self.metadata.columns:
      return [
          ontology.from_curie(curie) if curie is not None else None
          for curie in self.metadata['ontology_curie'].values
      ]
    else:
      return None

  # Track order / subset wrangling:
  def filter_tracks(self, mask: np.ndarray | list[bool]) -> 'JunctionData':
    """Filters tracks by a boolean mask.

    Args:
      mask: A boolean mask to select tracks.

    Returns:
      A new `JunctionData` object with the filtered tracks.
    """
    return JunctionData(
        junctions=self.junctions,
        values=self.values[:, mask],
        metadata=self.metadata.loc[mask],
        interval=self.interval,
        uns=self.uns,
    )

  def filter_to_strand(self, strand: str) -> 'JunctionData':
    """Filters junctions to a specific DNA strand.

    Args:
      strand: The strand to filter by ('+' or '-').

    Returns:
      A new `JunctionData` object with junctions on the specified strand.
    """
    mask = self.strands == strand
    return JunctionData(
        junctions=self.junctions[mask],
        values=self.values[mask, :],
        metadata=self.metadata,
        interval=self.interval,
        uns=self.uns,
    )

  def normalize_values(self, total_k: float = 10.0) -> 'JunctionData':
    """Normalizes the values by the k value."""
    values = self.values * total_k / self.values.sum()
    return JunctionData(
        junctions=self.junctions,
        values=values,
        metadata=self.metadata,
        interval=self.interval,
        uns=self.uns,
    )

  def filter_to_positive_strand(self) -> 'JunctionData':
    """Filters junctions to the positive DNA strand."""
    return self.filter_to_strand(genome.STRAND_POSITIVE)

  def filter_to_negative_strand(self) -> 'JunctionData':
    """Filters junctions to the negative DNA strand."""
    return self.filter_to_strand(genome.STRAND_NEGATIVE)

  def filter_by_tissue(self, tissue: str) -> 'JunctionData':
    """Filters tracks by GTEx tissue type.

    Args:
      tissue: The GTEx tissue type to filter by.

    Returns:
      A new `JunctionData` object with tracks from the specified tissue.

    Raises:
      ValueError: If the metadata does not contain a 'gtex_tissue' column.
    """
    if 'gtex_tissue' not in self.metadata.columns:
      raise ValueError(
          'Metadata does not contain gtex_tissue column. '
          f'Got {set(self.metadata.columns)}.'
      )
    return self.filter_tracks(self.metadata['gtex_tissue'] == tissue)

  def filter_by_name(self, name: str) -> 'JunctionData':
    """Filters tracks by name."""
    return self.filter_tracks(self.metadata['name'] == name)

  def filter_by_ontology(self, ontology_curie: str) -> 'JunctionData':
    """Filters tracks by ontology term.

    Args:
      ontology_curie: The ontology term CURIE to filter by.

    Returns:
      A new `JunctionData` object with tracks associated with the specified
        ontology term.

    Raises:
      ValueError: If the metadata does not contain an 'ontology_curie' column.
    """
    if 'ontology_curie' not in self.metadata.columns:
      raise ValueError(
          'Metadata does not contain ontology_curie column. '
          f'Got {set(self.metadata.columns)}.'
      )
    return self.filter_tracks(self.metadata['ontology_curie'] == ontology_curie)

  def intersect_with_interval(
      self, interval: genome.Interval
  ) -> 'JunctionData':
    """Returns the intersection of the junctions and the interval."""
    mask = np.array([j.overlaps(interval) for j in self.junctions])
    return JunctionData(
        junctions=self.junctions[mask],
        values=self.values[mask, :],
        metadata=self.metadata,
        interval=self.interval,
        uns=self.uns,
    )


def get_junctions_to_plot(
    *,
    predictions: JunctionData,
    name: str,
    strand: str,
    k_threshold: float | None = 0.0,
) -> list[genome.Junction]:
  """Gets a list of junctions to plot.

  Filters the junctions in the `predictions` by ontology term and strand, and
  applies a threshold on the `k` value (read count).

  Args:
    predictions: A `JunctionData` object containing junction predictions.
    name: The name to filter by.
    strand: The strand to filter by ('+' or '-').
    k_threshold: The minimum `k` value for a junction to be included. If None,
      use 5% of the maximum value.

  Returns:
    A list of `Junction` objects to plot.

  Raises:
    ValueError: If more than one track is found for the specified ontology term.
  """
  filtered = predictions.filter_by_name(name)
  if filtered.num_tracks > 1:
    raise ValueError(
        f'Expected only one ontology term, got {filtered.num_tracks}.'
    )
  if k_threshold is None:
    k_threshold = filtered.values.max() * 0.05
  # Round k for better visualization.
  filtered_junctions = []
  for interval, k in zip(filtered.junctions, filtered.values, strict=True):
    # Filter by threshold and strand.
    if interval.strand != strand:
      continue
    k = k.item()
    if k >= k_threshold:
      k = round(k, 2)

      filtered_junctions.append(
          genome.Junction(
              interval.chromosome,
              interval.start,
              interval.end,
              interval.strand,
              k=k,
          )
      )
  return filtered_junctions



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handling of biological ontologies."""

from collections.abc import Sequence
import dataclasses
import enum

from alphagenome.protos import dna_model_pb2
import immutabledict


class OntologyType(enum.IntEnum):
  """Supported ontology types.

  CLO: Cell Line Ontology.
  UBERON: Uber-anatomy ontology.
  CL: Cell Ontology.
  EFO: Experimental Factor Ontology.
  NTR: New Term Requested.
  """

  CLO = 1
  UBERON = 2
  CL = 3
  EFO = 4
  NTR = 5


_ONTOLOGY_TYPE_TO_PROTO_ENUM = immutabledict.immutabledict({
    OntologyType.CLO: dna_model_pb2.OntologyType.ONTOLOGY_TYPE_CLO,
    OntologyType.UBERON: dna_model_pb2.OntologyType.ONTOLOGY_TYPE_UBERON,
    OntologyType.CL: dna_model_pb2.OntologyType.ONTOLOGY_TYPE_CL,
    OntologyType.EFO: dna_model_pb2.OntologyType.ONTOLOGY_TYPE_EFO,
    OntologyType.NTR: dna_model_pb2.OntologyType.ONTOLOGY_TYPE_NTR,
})


@dataclasses.dataclass(frozen=True)
class OntologyTerm:
  """A single biological ontology term.

  Attributes:
    type: The ontology type.
    id: The ID of the term within the ontology.
  """

  type: OntologyType
  id: int

  @property
  def ontology_curie(self) -> str:
    """Returns the CURIE (Compact Uniform Resource Identifier) for the term."""
    return f'{self.type.name}:{self.id:07d}'

  def to_proto(self) -> dna_model_pb2.OntologyTerm:
    """Converts the ontology term to a protobuf message."""
    return dna_model_pb2.OntologyTerm(
        ontology_type=_ONTOLOGY_TYPE_TO_PROTO_ENUM[self.type], id=self.id
    )


def from_curie(ontology_curie: str) -> OntologyTerm:
  """Creates an `OntologyTerm` from a CURIE.

  Args:
    ontology_curie: The CURIE (Compact Uniform Resource Identifier) for the
      ontology term.

  Returns:
    An `OntologyTerm` object.
  """
  try:
    ontology_type, local_id = ontology_curie.split(':')
  except ValueError as e:
    raise ValueError(
        f'Invalid {ontology_curie=}. CURIEs must be of the form <type>:<id>.'
    ) from e
  try:
    ontology_type = OntologyType[ontology_type]
  except KeyError as e:
    raise ValueError(f'Cannot parse {ontology_type=}') from e
  return OntologyTerm(ontology_type, int(local_id))


def from_curies(ontology_curies: Sequence[str]) -> Sequence[OntologyTerm]:
  """Creates a list of `OntologyTerm` objects from a list of CURIEs.

  Args:
    ontology_curies: A list of CURIEs (Compact Uniform Resource Identifiers).

  Returns:
    A list of `OntologyTerm` objects.
  """
  return [from_curie(curie) for curie in ontology_curies]


def from_proto(proto: dna_model_pb2.OntologyTerm) -> OntologyTerm:
  """Creates an `OntologyTerm` from a protobuf message.

  Args:
    proto: An `OntologyTerm` protobuf message.

  Returns:
    An `OntologyTerm` object.
  """
  return OntologyTerm(
      OntologyType(proto.ontology_type),
      proto.id,
  )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from alphagenome.data import ontology
from alphagenome.data import track_data
from alphagenome.models import dna_output
import numpy as np
import pandas as pd


_EXAMPLE_ONTOLOGY_TERM = 'UBERON:0002037'

_ATAC_METADATA = pd.DataFrame(
    dict(
        name=['foo', 'bar', 'baz', 'buz', 'fux'],
        strand=['.'] * 5,
        padding=[False, False, False, False, True],
        ontology_curie=[
            _EXAMPLE_ONTOLOGY_TERM,
            _EXAMPLE_ONTOLOGY_TERM,
            _EXAMPLE_ONTOLOGY_TERM,
            'UBERON:0000005',
            None,
        ],
        extra=[1] * 5,
    )
)
_CAGE_METADATA = pd.DataFrame(
    dict(
        name=['foo', 'bar', 'baz'],
        strand=['.'] * 3,
        padding=[False] * 3,
        ontology_curie=[_EXAMPLE_ONTOLOGY_TERM] * 3,
        extra=[1] * 3,
    )
)
_DNASE_METADATA = pd.DataFrame(
    dict(name=['foo'], strand=['.'], padding=[False], extra=[1])
)
_RNA_SEQ_METADATA = pd.DataFrame(
    dict(
        name=['foo', 'foo', 'baz', 'buz'],
        strand=['+', '-', '.', '.'],
        ontology_curie=[_EXAMPLE_ONTOLOGY_TERM] * 4,
        padding=[False] * 4,
        extra=[1] * 4,
    )
)
_CHIP_HISTONE_METADATA = pd.DataFrame(
    dict(
        name=['foo', 'bar', 'baz', 'buz', 'quz'],
        strand=['.'] * 5,
        padding=[False] * 5,
        extra=[1] * 5,
    )
)
_SPLICING_METADATA = pd.DataFrame(
    dict(
        name=['foo', 'bar', 'foo', 'bar'],
        strand=['+'] * 2 + ['-'] * 2,
        padding=[False] * 4,
        extra=[1] * 4,
    )
)


def _create_track_data(
    sequence_length: int,
    metadata: pd.DataFrame,
    *,
    ontology_terms: list[ontology.OntologyTerm | None] | None = None,
) -> track_data.TrackData:
  metadata = metadata[~metadata['padding']]
  if ontology_terms is not None:
    if 'ontology_curie' in metadata.columns:
      metadata = metadata[
          metadata['ontology_curie'].isin(
              [o.ontology_curie for o in ontology_terms if o is not None]
          )
      ]
    else:
      metadata = metadata[slice(0, 0)]
  return track_data.TrackData(
      np.zeros(
          (sequence_length, len(metadata)),
          dtype=np.float32,
      ),
      metadata,
  )


def _get_output(sequence_length: int = 10):
  return dna_output.Output(
      atac=_create_track_data(sequence_length, _ATAC_METADATA),
      cage=_create_track_data(sequence_length, _CAGE_METADATA),
      dnase=_create_track_data(sequence_length, _DNASE_METADATA),
      rna_seq=_create_track_data(sequence_length, _RNA_SEQ_METADATA),
      chip_histone=None,
      chip_tf=None,
      splice_sites=None,
      splice_site_usage=None,
      splice_junctions=None,
      contact_maps=None,
  )


def _get_rnaseq_atac_output():
  """Manually specified RNA-seq and ATAC Outputs for testing diff and add."""
  atac_track_data_1 = track_data.TrackData(
      values=np.array(
          [
              [0.0, 0.2, 0.3, 0.4],
              [0.1, 0.3, 0.5, 0.6],
              [0.2, 0.4, 0.7, 0.8],
              [0.5, 0.1, 0.2, 0.6],
              [0.2, 0.4, 0.7, 0.8],
          ],
          dtype=np.float32,
      ),
      metadata=pd.DataFrame({
          'name': ['track_0', 'track_1', 'track_2', 'track_3'],
          'strand': '.',
          'tissue': ['HEART', 'BRAIN', 'BANANA', 'STOMACH'],
          'padding': [False, False, False, False],
      }),
  )

  atac_track_data_2 = track_data.TrackData(
      values=-1 * atac_track_data_1.values,
      metadata=atac_track_data_1.metadata,
  )

  rna_seq_track_data_1 = track_data.TrackData(
      values=np.array(
          [
              [1.0, 1.1, 1.2, 1.3, 1.4],
              [1.5, 1.6, 1.7, 1.8, 1.9],
              [1.2, 1.4, 1.7, 1.8, 1.9],
              [1.3, 1.4, 1.7, 1.8, 1.9],
              [2.0, 2.1, 2.2, 2.3, 2.4],
          ],
          dtype=np.float32,
      ),
      metadata=pd.DataFrame({
          'name': ['track_0', 'track_1', 'track_2', 'track_3', 'track_4'],
          'strand': ['+', '-', '-', '+', '.'],
          'tissue': ['BLADDER', 'PANCREAS', 'SKIN', 'PANCREAS', ''],
          'padding': [False, False, False, False, True],
      }),
  )

  rna_seq_track_data_2 = track_data.TrackData(
      values=-1 * rna_seq_track_data_1.values,
      metadata=rna_seq_track_data_1.metadata,
  )

  # Build the 2 Output objects.
  output_1 = dna_output.Output(
      atac=atac_track_data_1,
      rna_seq=rna_seq_track_data_1,
  )
  output_2 = dna_output.Output(
      atac=atac_track_data_2,
      rna_seq=rna_seq_track_data_2,
  )
  output_3 = dna_output.Output(
      atac=atac_track_data_2,
      rna_seq=None,
  )
  return output_1, output_2, output_3


class OutputTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.output = _get_output()
    # Additional manually specified outputs for testing diff and add.
    self.output1, self.output2, self.output3 = _get_rnaseq_atac_output()

  def test_get(self):
    atac_output = self.output.get(dna_output.OutputType.ATAC)
    self.assertIsNotNone(atac_output)
    self.assertLen(
        _ATAC_METADATA[~_ATAC_METADATA['padding']],
        atac_output.num_tracks,
    )
    self.assertIsNone(self.output.get(dna_output.OutputType.CONTACT_MAPS))

  def test_map_track_data(self):

    def _fn(tdata, output_type):
      del output_type
      return tdata.select_tracks_by_index([0])

    filtered_output = self.output.map_track_data(_fn)

    atac_output = filtered_output.get(dna_output.OutputType.ATAC)
    self.assertIsNotNone(atac_output)
    self.assertEqual(atac_output.num_tracks, 1)

    cage_output = filtered_output.get(dna_output.OutputType.CAGE)
    self.assertIsNotNone(cage_output)
    self.assertEqual(cage_output.num_tracks, 1)

    dnase_output = filtered_output.get(dna_output.OutputType.DNASE)
    self.assertIsNotNone(dnase_output)
    self.assertEqual(dnase_output.num_tracks, 1)

    rna_seq_output = filtered_output.get(dna_output.OutputType.RNA_SEQ)
    self.assertIsNotNone(rna_seq_output)
    self.assertEqual(rna_seq_output.num_tracks, 1)

    self.assertIsNone(self.output.get(dna_output.OutputType.CONTACT_MAPS))

  def test_filter_to_strand_plus(self):
    filtered_output = self.output.filter_to_strand('+')

    atac_output = filtered_output.get(dna_output.OutputType.ATAC)
    self.assertIsNotNone(atac_output)
    self.assertEqual(atac_output.num_tracks, 0)

    cage_output = filtered_output.get(dna_output.OutputType.CAGE)
    self.assertIsNotNone(cage_output)
    self.assertEqual(cage_output.num_tracks, 0)

    dnase_output = filtered_output.get(dna_output.OutputType.DNASE)
    self.assertIsNotNone(dnase_output)
    self.assertEqual(dnase_output.num_tracks, 0)

    rna_seq_output = filtered_output.get(dna_output.OutputType.RNA_SEQ)
    self.assertIsNotNone(rna_seq_output)
    self.assertEqual(rna_seq_output.num_tracks, 1)

    self.assertIsNone(filtered_output.get(dna_output.OutputType.CONTACT_MAPS))

  def test_filter_to_strand_minus(self):
    filtered_output = self.output.filter_to_strand('-')

    atac_output = filtered_output.get(dna_output.OutputType.ATAC)
    self.assertIsNotNone(atac_output)
    self.assertEqual(atac_output.num_tracks, 0)

    cage_output = filtered_output.get(dna_output.OutputType.CAGE)
    self.assertIsNotNone(cage_output)
    self.assertEqual(cage_output.num_tracks, 0)

    dnase_output = filtered_output.get(dna_output.OutputType.DNASE)
    self.assertIsNotNone(dnase_output)
    self.assertEqual(dnase_output.num_tracks, 0)

    rna_seq_output = filtered_output.get(dna_output.OutputType.RNA_SEQ)
    self.assertIsNotNone(rna_seq_output)
    self.assertEqual(rna_seq_output.num_tracks, 1)

    self.assertIsNone(filtered_output.get(dna_output.OutputType.CONTACT_MAPS))

  def test_filter_to_strand_unstranded(self):
    filtered_output = self.output.filter_to_strand('.')

    atac_output = filtered_output.get(dna_output.OutputType.ATAC)
    self.assertIsNotNone(atac_output)
    self.assertLen(
        _ATAC_METADATA[~_ATAC_METADATA['padding']], atac_output.num_tracks
    )

    cage_output = filtered_output.get(dna_output.OutputType.CAGE)
    self.assertIsNotNone(cage_output)
    self.assertLen(
        _CAGE_METADATA[~_CAGE_METADATA['padding']],
        cage_output.num_tracks,
    )

    dnase_output = filtered_output.get(dna_output.OutputType.DNASE)
    self.assertIsNotNone(dnase_output)
    self.assertLen(
        _DNASE_METADATA[~_DNASE_METADATA['padding']],
        dnase_output.num_tracks,
    )

    rna_seq_output = filtered_output.get(dna_output.OutputType.RNA_SEQ)
    self.assertIsNotNone(rna_seq_output)
    self.assertEqual(rna_seq_output.num_tracks, 2)

    self.assertIsNone(filtered_output.get(dna_output.OutputType.CONTACT_MAPS))

  def test_filter_output_type(self):
    output = self.output.filter_output_type({dna_output.OutputType.ATAC})
    self.assertIsInstance(
        output.get(dna_output.OutputType.ATAC), track_data.TrackData
    )
    self.assertIsNone(output.get(dna_output.OutputType.CAGE))
    self.assertIsNone(output.get(dna_output.OutputType.DNASE))
    self.assertIsNone(output.get(dna_output.OutputType.CONTACT_MAPS))

  def test_filter_ontology_terms(self):
    filtered_output = self.output.filter_ontology_terms(
        {ontology.from_curie(_EXAMPLE_ONTOLOGY_TERM)}
    )
    dnase_output = filtered_output.get(dna_output.OutputType.DNASE)
    self.assertIsNotNone(dnase_output)
    self.assertLen(
        _DNASE_METADATA,
        dnase_output.num_tracks,
    )
    atac_output = filtered_output.get(dna_output.OutputType.ATAC)
    self.assertIsNotNone(atac_output)
    self.assertEqual(
        atac_output.num_tracks,
        np.sum(_ATAC_METADATA['ontology_curie'] == _EXAMPLE_ONTOLOGY_TERM),
    )

  def test_output_diff(self):
    output_diff = self.output1 - self.output2
    self.assertIsInstance(output_diff.rna_seq, track_data.TrackData)
    self.assertIsInstance(output_diff.atac, track_data.TrackData)

    # To fix failing build (since these fields can be None). Can't be list comp.
    assert self.output1.rna_seq is not None
    assert self.output2.rna_seq is not None
    assert output_diff.rna_seq is not None
    assert self.output1.atac is not None
    assert self.output2.atac is not None
    assert output_diff.atac is not None

    manual_diff_rnaseq = (
        self.output1.rna_seq.values - self.output2.rna_seq.values
    )
    np.testing.assert_array_equal(
        output_diff.rna_seq.values, manual_diff_rnaseq
    )

    manual_diff_atac = self.output1.atac.values - self.output2.atac.values
    np.testing.assert_array_equal(output_diff.atac.values, manual_diff_atac)

    # output3 has None for rna_seq. Where None is passed, just return None
    # rather than trying to do arithmetic.
    output_diff = self.output1 - self.output3
    self.assertIsNone(output_diff.rna_seq)
    self.assertIsNotNone(output_diff.atac)

    output_diff = self.output3 - self.output1
    self.assertIsNone(output_diff.rna_seq)
    self.assertIsNotNone(output_diff.atac)

  def test_output_add(self):
    output_sum = self.output1 + self.output2
    self.assertIsInstance(output_sum.rna_seq, track_data.TrackData)
    self.assertIsInstance(output_sum.atac, track_data.TrackData)

    # To fix failing build (since these fields can be None). Can't be list comp.
    assert self.output1.rna_seq is not None
    assert self.output2.rna_seq is not None
    assert output_sum.rna_seq is not None
    assert self.output1.atac is not None
    assert self.output2.atac is not None
    assert output_sum.atac is not None

    manual_sum_rnaseq = (
        self.output1.rna_seq.values + self.output2.rna_seq.values
    )
    np.testing.assert_array_equal(output_sum.rna_seq.values, manual_sum_rnaseq)

    manual_sum_atac = self.output1.atac.values + self.output2.atac.values
    np.testing.assert_array_equal(output_sum.atac.values, manual_sum_atac)

    # output3 has None for rna_seq. Where None is passed, just return None
    # rather than trying to do arithmetic.
    output_sum = self.output1 + self.output3
    self.assertIsNone(output_sum.rna_seq)
    self.assertIsNotNone(output_sum.atac)

    output_sum = self.output3 + self.output1
    self.assertIsNone(output_sum.rna_seq)
    self.assertIsNotNone(output_sum.atac)

  def test_output_metadata_from_outputs(self):
    output = _get_output()
    mapping = {
        dna_output.OutputType.ATAC: output.atac,
        dna_output.OutputType.CAGE: output.cage,
        dna_output.OutputType.DNASE: output.dnase,
        dna_output.OutputType.RNA_SEQ: output.rna_seq,
    }
    output_metadata = dna_output.OutputMetadata.from_outputs(mapping)
    for output_type in dna_output.OutputType:
      if output_type in mapping:
        self.assertIs(
            output_metadata.get(output_type),
            output.get(output_type).metadata,  # pytype: disable=attribute-error
        )
      else:
        self.assertIsNone(output_metadata.get(output_type))

  def test_output_metadata_concatenate(self):
    output_metadata = dna_output.OutputMetadata(
        atac=track_data.TrackMetadata(
            {'name': [f'track_{i}' for i in range(32)], 'strand': ['.'] * 32}
        ),
        dnase=track_data.TrackMetadata(
            {'name': [f'track_{i}' for i in range(10)], 'strand': ['.'] * 10}
        ),
    )
    df = output_metadata.concatenate()
    self.assertLen(df, 42)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script to process GTF into feather file."""

import os
import tempfile
from urllib import parse
from urllib import request

from absl import app
from absl import flags
from absl import logging
import pyranges


_GTF_PATH = flags.DEFINE_string(
    'gtf_path', None, 'Path to GTF file.', required=True
)
_OUTPUT_PATH = flags.DEFINE_string(
    'output_path', None, 'Path to output feather file.', required=True
)


def main(_) -> None:
  logging.info('Reading GTF from %s', _GTF_PATH.value)
  url = parse.urlparse(_GTF_PATH.value)
  if all([url.scheme, url.netloc]):
    with tempfile.TemporaryDirectory() as d:
      path, _ = request.urlretrieve(
          _GTF_PATH.value,
          filename=os.path.join(d, os.path.basename(_GTF_PATH.value)),
      )
      logging.info('Downloaded GTF to %s', path)
      gtf = pyranges.read_gtf(path, as_df=True, duplicate_attr=True)
  else:
    gtf = pyranges.read_gtf(_GTF_PATH.value, as_df=True, duplicate_attr=True)

  gtf['gene_id_nopatch'] = gtf['gene_id'].str.split('.', expand=True)[0]

  logging.info('Writing GTF to %s', _OUTPUT_PATH.value)
  gtf.to_feather(_OUTPUT_PATH.value)


if __name__ == '__main__':
  app.run(main)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for AlphaGenome model outputs."""

from collections.abc import Callable, Iterable, Mapping
import dataclasses
import enum
from typing import Literal

from alphagenome import typing
from alphagenome.data import junction_data
from alphagenome.data import ontology
from alphagenome.data import track_data
from alphagenome.protos import dna_model_pb2
import pandas as pd


class OutputType(enum.Enum):
  """Enumeration of all the available types of outputs.

  Attributes:
    ATAC: ATAC-seq tracks capturing chromatin accessibility.
    CAGE: CAGE (Cap Analysis of Gene Expression) tracks capturing gene
      expression.
    DNASE: DNase I hypersensitive site tracks capturing chromatin accessibility.
    RNA_SEQ: RNA sequencing tracks capturing gene expression.
    CHIP_HISTONE: ChIP-seq tracks capturing histone modifications.
    CHIP_TF: ChIP-seq tracks capturing transcription factor binding.
    SPLICE_SITES: Splice site tracks capturing donor and acceptor splice sites.
    SPLICE_SITE_USAGE: Splice site usage tracks capturing the fraction of the
      time that each splice site is used.
    SPLICE_JUNCTIONS: Splice junction tracks capturing split read RNA-seq counts
      for each junction.
    CONTACT_MAPS: Contact map tracks capturing 3D DNA-DNA contact probabilities.
    PROCAP: Precision Run-On sequencing and capping, used to measure gene
      expression.
  """

  ATAC = dna_model_pb2.OUTPUT_TYPE_ATAC
  CAGE = dna_model_pb2.OUTPUT_TYPE_CAGE
  DNASE = dna_model_pb2.OUTPUT_TYPE_DNASE
  RNA_SEQ = dna_model_pb2.OUTPUT_TYPE_RNA_SEQ
  CHIP_HISTONE = dna_model_pb2.OUTPUT_TYPE_CHIP_HISTONE
  CHIP_TF = dna_model_pb2.OUTPUT_TYPE_CHIP_TF
  SPLICE_SITES = dna_model_pb2.OUTPUT_TYPE_SPLICE_SITES
  SPLICE_SITE_USAGE = dna_model_pb2.OUTPUT_TYPE_SPLICE_SITE_USAGE
  SPLICE_JUNCTIONS = dna_model_pb2.OUTPUT_TYPE_SPLICE_JUNCTIONS
  CONTACT_MAPS = dna_model_pb2.OUTPUT_TYPE_CONTACT_MAPS
  PROCAP = dna_model_pb2.OUTPUT_TYPE_PROCAP

  def __lt__(self, other: 'OutputType'):
    """Compares if an other `OutputType` enum value is less than this one."""
    return self.value < other.value

  def to_proto(self) -> dna_model_pb2.OutputType:
    """Converts the `OutputType` enum to a protobuf enum."""
    return self.value

  def __repr__(self) -> str:
    """Returns name of the `OutputType` enum as the string representation."""
    return self.name


@typing.jaxtyped
@dataclasses.dataclass(frozen=True)
class Output:
  """Model outputs for a single prediction.

  Attributes:
    atac: TrackData of type OutputType.ATAC.
    cage: TrackData of type OutputType.CAGE.
    dnase: TrackData of type OutputType.DNASE.
    rna_seq: TrackData of type OutputType.RNA_SEQ.
    chip_histone: TrackData of type OutputType.CHIP_HISTONE.
    chip_tf: TrackData of type OutputType.CHIP_TF.
    splice_sites: TrackData of type OutputType.SPLICE_SITES.
    splice_site_usage: TrackData of type OutputType.SPLICE_SITE_USAGE.
    splice_junctions: TrackData of type OutputType.SPLICE_JUNCTIONS.
    contact_maps: TrackData of type OutputType.CONTACT_MAPS.
    procap: TrackData of type OutputType.PROCAP.
  """

  atac: track_data.TrackData | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.ATAC}
  )
  cage: track_data.TrackData | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.CAGE}
  )
  dnase: track_data.TrackData | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.DNASE}
  )
  rna_seq: track_data.TrackData | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.RNA_SEQ}
  )
  chip_histone: track_data.TrackData | None = dataclasses.field(
      default=None,
      metadata={'output_type': OutputType.CHIP_HISTONE},
  )
  chip_tf: track_data.TrackData | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.CHIP_TF}
  )

  splice_sites: track_data.TrackData | None = dataclasses.field(
      default=None,
      metadata={'output_type': OutputType.SPLICE_SITES},
  )
  splice_site_usage: track_data.TrackData | None = dataclasses.field(
      default=None,
      metadata={'output_type': OutputType.SPLICE_SITE_USAGE},
  )
  splice_junctions: junction_data.JunctionData | None = dataclasses.field(
      default=None,
      metadata={'output_type': OutputType.SPLICE_JUNCTIONS},
  )

  contact_maps: track_data.TrackData | None = dataclasses.field(
      default=None,
      metadata={'output_type': OutputType.CONTACT_MAPS},
  )
  procap: track_data.TrackData | None = dataclasses.field(
      default=None,
      metadata={'output_type': OutputType.PROCAP},
  )

  def get(
      self, output_type: OutputType
  ) -> track_data.TrackData | junction_data.JunctionData | None:
    """Gets the track data for the specified output type.

    Args:
      output_type: The type of output to retrieve.

    Returns:
      The track data for the specified output type, or None if no such data
      exists.
    """
    for field in dataclasses.fields(self):
      if field.metadata['output_type'] == output_type:
        return getattr(self, field.name)

  def map_track_data(
      self,
      fn: Callable[
          [track_data.TrackData, OutputType],
          track_data.TrackData | None,
      ],
  ) -> 'Output':
    """Applies a transformation function to each `TrackData`.

    Args:
      fn: The function to apply to each `TrackData`. It should take a
        `TrackData` object and an `OutputType` enum as input and return a
        `TrackData` object or None.

    Returns:
      A new `Output` object with the transformed track data.
    """
    output_dict = {}
    for field in dataclasses.fields(self):
      output_type = field.metadata['output_type']
      value = self.get(output_type)
      if isinstance(value, track_data.TrackData):
        output_dict[field.name] = fn(value, output_type)
      else:
        output_dict[field.name] = value
    return Output(**output_dict)

  def filter_to_strand(self, strand: Literal['+', '-', '.']) -> 'Output':
    """Filters tracks by DNA strand.

    Args:
      strand: The strand to filter by ('+', '-', or '.').

    Returns:
      A new `Output` object with only the tracks on the specified strand.
    """

    def _filter_to_strand(
        tdata: track_data.TrackData, output_type: OutputType
    ) -> track_data.TrackData | None:
      del output_type  # Unused.
      return tdata.filter_tracks(tdata.strands == strand)

    return self.map_track_data(_filter_to_strand)

  def filter_ontology_terms(
      self, ontology_terms: Iterable[ontology.OntologyTerm]
  ) -> 'Output':
    """Filters tracks to specific ontology terms.

    Args:
      ontology_terms: An iterable of `OntologyTerm` objects to filter to.

    Returns:
      A new `Output` object with only the tracks associated with the specified
      ontology terms.
    """

    def _filter_ontology(
        tdata: track_data.TrackData, output_type: OutputType
    ) -> track_data.TrackData | None:
      del output_type  # Unused.
      if track_ontologies := tdata.ontology_terms:
        return tdata.filter_tracks(
            [o in ontology_terms for o in track_ontologies]
        )
      else:
        return tdata

    return self.map_track_data(_filter_ontology)

  def filter_output_type(self, output_types: Iterable[OutputType]) -> 'Output':
    """Filters tracks to specific output type.

    Args:
      output_types: An iterable of `OutputType` enums to filter by.

    Returns:
      A new `Output` object with only the tracks of the specified output types.
    """
    output_dict = {}
    for field in dataclasses.fields(self):
      output_type = field.metadata['output_type']
      if output_type in output_types:
        output_dict[field.name] = self.get(output_type)
      else:
        output_dict[field.name] = None

    return Output(**output_dict)

  def resize(self, width: int) -> 'Output':
    """Resizes all track data to a specified width.

    Args:
      width: The desired width in base pairs.

    Returns:
      A new `Output` object with resized track data.
    """
    return self.map_track_data(lambda tdata, _: tdata.resize(width))

  def __add__(self, other: 'Output') -> 'Output':
    """Adds the values of two `Output` objects element-wise.

    Args:
      other: The `Output` object to add.

    Returns:
      A new `Output` object with the summed values.
    """

    def add_track_data(
        track_data1,
        output_type: OutputType,
    ):
      """Adds two `TrackData` objects for a specific output type."""
      track_data2 = other.get(output_type)
      if track_data1 is None or track_data2 is None:
        return None
      return track_data1 + track_data2

    return self.map_track_data(add_track_data)

  def __sub__(self, other: 'Output') -> 'Output':
    """Subtracts the values of two `Output` objects element-wise.

    Args:
      other: The `Output` object to subtract.

    Returns:
      A new `Output` object with the difference of the values.
    """

    def sub_track_data(track_data1, output_type: OutputType):
      """Subtracts two `TrackData` objects for a specific output type."""
      track_data2 = other.get(output_type)
      if track_data1 is None or track_data2 is None:
        return None
      return track_data1 - track_data2

    return self.map_track_data(sub_track_data)


@dataclasses.dataclass(frozen=True, kw_only=True)
class OutputMetadata:
  """Metadata detailing the content of model output.

  Attributes:
    atac: Metadata for ATAC-seq tracks.
    cage: Metadata for CAGE tracks.
    dnase: Metadata for DNase I hypersensitive site tracks.
    rna_seq: Metadata for RNA sequencing tracks.
    chip_histone: Metadata for ChIP-seq tracks capturing histone modifications.
    chip_tf: Metadata for ChIP-seq tracks capturing transcription factor
      binding.
    splice_sites: Metadata for splice site tracks.
    splice_site_usage: Metadata for splice site usage tracks.
    splice_junctions: Metadata for splice junction tracks.
    contact_maps: Metadata for contact map tracks.
    procap: Metadata for procap tracks.
  """

  atac: track_data.TrackMetadata | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.ATAC}
  )
  cage: track_data.TrackMetadata | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.CAGE}
  )
  dnase: track_data.TrackMetadata | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.DNASE}
  )
  rna_seq: track_data.TrackMetadata | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.RNA_SEQ}
  )
  chip_histone: track_data.TrackMetadata | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.CHIP_HISTONE}
  )
  chip_tf: track_data.TrackMetadata | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.CHIP_TF}
  )
  splice_sites: track_data.TrackMetadata | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.SPLICE_SITES}
  )
  splice_site_usage: track_data.TrackMetadata | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.SPLICE_SITE_USAGE}
  )
  splice_junctions: junction_data.JunctionMetadata | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.SPLICE_JUNCTIONS}
  )
  contact_maps: track_data.TrackMetadata | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.CONTACT_MAPS}
  )
  procap: track_data.TrackMetadata | None = dataclasses.field(
      default=None, metadata={'output_type': OutputType.PROCAP}
  )

  def get(self, output: OutputType) -> track_data.TrackMetadata | None:
    """Gets the track metadata for a given output type.

    Args:
      output: The `OutputType` enum value.

    Returns:
      The corresponding track metadata, or None if it doesn't exist.
    """
    match output:
      case OutputType.ATAC:
        return self.atac
      case OutputType.CAGE:
        return self.cage
      case OutputType.DNASE:
        return self.dnase
      case OutputType.RNA_SEQ:
        return self.rna_seq
      case OutputType.CHIP_HISTONE:
        return self.chip_histone
      case OutputType.CHIP_TF:
        return self.chip_tf
      case OutputType.SPLICE_SITES:
        return self.splice_sites
      case OutputType.SPLICE_JUNCTIONS:
        return self.splice_junctions
      case OutputType.SPLICE_SITE_USAGE:
        return self.splice_site_usage
      case OutputType.CONTACT_MAPS:
        return self.contact_maps
      case OutputType.PROCAP:
        return self.procap

  @classmethod
  def from_outputs(
      cls,
      outputs: Mapping[
          OutputType, track_data.TrackData | junction_data.JunctionData
      ],
  ) -> 'OutputMetadata':
    """Creates an `OutputMetadata` from a mapping of output types to data.

    Args:
      outputs: A mapping from `OutputType` to `TrackData` or `JunctionData`.

    Returns:
      An `OutputMetadata` object.
    """
    kwargs = {}
    for field in dataclasses.fields(cls):
      output_type = field.metadata['output_type']
      if data := outputs.get(output_type):
        kwargs[field.name] = data.metadata
    return cls(**kwargs)

  def concatenate(self) -> track_data.TrackMetadata:
    """Concatenates all metadata into a single DataFrame.

    Returns:
      A pandas DataFrame containing all the track metadata, with an additional
      column 'output_type' specifying the type of each track.
    """
    df_list = []
    for output_type in OutputType:
      if (df := self.get(output_type)) is not None:
        assert isinstance(df, pd.DataFrame)
        df_list.append(df.assign(output_type=output_type))
    return pd.concat(df_list)


@typing.jaxtyped
@dataclasses.dataclass(frozen=True)
class VariantOutput:
  """Model outputs for a variant prediction.

  Attributes:
    reference: The model output for the reference sequence.
    alternate: The model output for the alternate sequence.
  """

  reference: Output
  alternate: Output



# Copyright 2025 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Benchmark for tensor_utils."""

import functools
import itertools
import tracemalloc

from absl import flags
from alphagenome import tensor_utils
from alphagenome.protos import tensor_pb2
import google_benchmark
import numpy as np

_MB_TO_BYTES = 2**20

_TENSOR_SIZE_IN_MB = flags.DEFINE_integer(
    'tensor_size_in_mb', 128, 'Size of the tensor to benchmark.'
)

_COMPRESSION_TYPE = flags.DEFINE_enum(
    'compression_type',
    'zstd',
    ['none', 'zstd'],
    'Compression type to benchmark.',
)


@google_benchmark.register
@google_benchmark.option.use_real_time()
@google_benchmark.option.unit(google_benchmark.kMillisecond)
@google_benchmark.option.range_multiplier(2)
@google_benchmark.option.range(1, limit=10)
def pack_tensor_benchmark_chunks(state):
  """Benchmark for packing tensors."""
  rng = np.random.default_rng(seed=42)
  num_elements = (_TENSOR_SIZE_IN_MB.value * _MB_TO_BYTES) // np.dtype(
      np.float32
  ).itemsize
  values = rng.random(num_elements, dtype=np.float32)
  compression_type = tensor_pb2.CompressionType.Value(
      f'COMPRESSION_TYPE_{_COMPRESSION_TYPE.value.upper()}'
  )

  uncompressed_bytes = 0
  compressed_bytes = 0
  tracemalloc.start()

  while state:
    packed, chunks = tensor_utils.pack_tensor(
        values,
        bytes_per_chunk=state.range(0) * _MB_TO_BYTES,
        compression_type=compression_type,
    )
    uncompressed_bytes += values.nbytes
    compressed_bytes += functools.reduce(
        lambda x, y: x + len(y.data), itertools.chain(chunks, [packed.array]), 0
    )

  _, peak_mem = tracemalloc.get_traced_memory()
  tracemalloc.stop()

  state.counters['compressed_bytes'] = google_benchmark.Counter(
      compressed_bytes
  )
  state.counters['uncompressed_bytes'] = google_benchmark.Counter(
      uncompressed_bytes
  )
  state.counters['peak_mem'] = google_benchmark.Counter(peak_mem)


@google_benchmark.register
@google_benchmark.option.use_real_time()
@google_benchmark.option.unit(google_benchmark.kMillisecond)
def unpack_tensor_benchmark_chunks(state):
  """Benchmark for unpacking tensors."""
  rng = np.random.default_rng(seed=42)
  num_elements = (_TENSOR_SIZE_IN_MB.value * _MB_TO_BYTES) // np.dtype(
      np.float32
  ).itemsize
  compression_type = tensor_pb2.CompressionType.Value(
      f'COMPRESSION_TYPE_{_COMPRESSION_TYPE.value.upper()}'
  )
  packed, chunks = tensor_utils.pack_tensor(
      rng.random(num_elements, dtype=np.float32),
      bytes_per_chunk=1 * _MB_TO_BYTES,
      compression_type=compression_type,
  )

  while state:
    tensor_utils.unpack_proto(packed, chunks)


if __name__ == '__main__':
  google_benchmark.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configuration file for the Sphinx documentation builder."""

#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import importlib.metadata
import inspect
import os
import sys

# The package is installed by Readthedocs before sphinx building.
import alphagenome  # pylint: disable=unused-import, g-import-not-at-top
import alphagenome.models.dna_client  # pylint: disable=unused-import, g-import-not-at-top

# -- Project information -----------------------------------------------------

project = 'alphagenome'
project_info = importlib.metadata.metadata(project)
author = project_info['Author']
copyright = f'2024, {author}'  # pylint: disable=redefined-builtin
version = project_info['Version']
repository_url = f'https://github.com/google-deepmind/{project}'


# The full version, including alpha/beta/rc tags
release = version
# Warn if links are broken
nitpicky = True

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'myst_nb',
    'sphinx_design',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinxcontrib.bibtex',
    'sphinx_autodoc_typehints',
    'sphinx.ext.mathjax',
    'IPython.sphinxext.ipython_console_highlighting',
    'sphinx.ext.coverage',
    'sphinx_copybutton',
    'sphinx_remove_toctrees',
    'sphinx.ext.linkcode',
]

autosummary_generate = True
autodoc_member_order = 'groupwise'
default_role = 'literal'
bibtex_reference_style = 'author_year'
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_use_rtype = True
napoleon_use_param = True
myst_heading_anchors = 6  # Create heading anchors for h1-h6
autodoc_mock_imports = [
    'google.protobuf.runtime_version',
    'google.protobuf.internal.builder',
    'absl',
    'alphagenome.protos',
]
remove_from_toctrees = ['api/generated/*']
bibtex_bibfiles = ['refs.bib']

myst_enable_extensions = [
    'amsmath',
    'colon_fence',
    'deflist',
    'dollarmath',
    'html_image',
    'html_admonition',
    'attrs_inline',
    'attrs_block',
]

# TODO(b/372225132): Resolve showing notebook output without executing.
# TODO(b/372226231): Resolve not modifying notebook when building docs.
myst_url_schemes = ['http', 'https', 'mailto']
nb_output_stderr = 'remove'
nb_execution_mode = 'off'
nb_merge_streams = True
typehints_defaults = 'braces'

source_suffix = {
    '.rst': 'restructuredtext',
    '.ipynb': 'myst-nb',
    '.myst': 'myst-nb',
}

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'anndata': ('https://anndata.readthedocs.io/en/stable/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'jax': ('https://jax.readthedocs.io/en/latest/', None),
    'pandas': ('https://pandas.pydata.org/docs/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'protos']

# -- Options for autodoc -----------------------------------------------------

autodoc_default_options = {
    'member-order': 'bysource',
    'special-members': True,
    'exclude-members': '__repr__, __str__, __weakref__',
}


# -- Source code links -----------------------------------------------------


def linkcode_resolve(domain, info):
  """Resolve a GitHub URL corresponding to Python object."""
  if domain != 'py':
    return None

  try:
    mod = sys.modules[info['module']]
  except ImportError:
    return None

  obj = mod
  try:
    for attr in info['fullname'].split('.'):
      obj = getattr(obj, attr)
  except AttributeError:
    return None
  else:
    obj = inspect.unwrap(obj)

  try:
    filename = inspect.getsourcefile(obj)
  except TypeError:
    return None

  try:
    source, lineno = inspect.getsourcelines(obj)
  except OSError:
    return None

  path = os.path.relpath(filename, start=os.path.dirname(alphagenome.__file__))
  return (
      f'{repository_url}/tree/main/src/{project}/'
      f'{path}#L{lineno}#L{lineno + len(source) - 1}'
  )


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'sphinx_book_theme'
html_title = 'AlphaGenome'
pygments_style = 'default'
html_theme_options = {
    'repository_url': repository_url,
    'repository_branch': 'main',
    'use_repository_button': True,
    'launch_buttons': {
        'colab_url': 'https://colab.research.google.com',
    },
    'article_header_start': ['toggle-primary-sidebar.html', 'breadcrumbs'],
    'show_prev_next': False,
}
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# TODO: b/377291190 - Look at adding notebook support (see haiku example)
